from __future__ import annotations
import argparse
import sys
from typing import Any, Optional

from evaluation_system import __version__
from evaluation_system.misc import logger
import freva
from .utils import BaseParser

CLI = "HistoryCli"


class HistoryCli(BaseParser):
    """Class that constructs the History Query Parser."""

    desc = "Read the plugin application history."

    def __init__(
        self,
        command: str = "freva",
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the history sub arg. parser."""
        subparser = parser or argparse.ArgumentParser(
            prog=f"{command}-history",
            description=self.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        subparser.add_argument(
            "--full-text",
            default=False,
            action="store_true",
            help="Show the complete configuration.",
        )
        subparser.add_argument(
            "--return-command",
            default=False,
            action="store_true",
            help=(
                "Show freva commands belonging to the history entries "
                "instead of the entries themself."
            ),
        )
        subparser.add_argument(
            "--limit",
            default=10,
            type=int,
            help="Limit the number of displayed entries to N",
        )
        subparser.add_argument(
            "--plugin",
            default=None,
            type=str,
            help="Display only entries of selected plugin.",
        )
        subparser.add_argument(
            "--since",
            default=None,
            type=str,
            help="Retrieve entries older than date",
        )
        subparser.add_argument(
            "--until",
            default=None,
            type=str,
            help="Retrieve entries newer than date",
        )
        subparser.add_argument(
            "--entry-ids",
            default=None,
            type=str,
            nargs="+",
            help="Select entry id(s) (e.g. --entry-ids 1 --entry-ids 2 )",
        )
        subparser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser = subparser
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace, other_args: Optional[list[str]], **kwargs: Any
    ) -> None:
        """Call the databrowser command and print the results."""
        kwargs.pop("other_args", "")
        try:
            if len(kwargs["entry_ids"]) == 1:
                kwargs["entry_ids"] = kwargs["entry_ids"][0].split(",")
        except TypeError:
            pass
        commands = freva.history(_return_dict=False, **kwargs)
        if not commands:
            logger.error("No results. Check query.")
            return
        if args.return_command:
            result = "\n".join(commands)
        else:
            result = "\n".join(
                [c.__str__(compact=not args.full_text) for c in commands]
            )
        print(result)


def main(argv: Optional[list[str]] = None) -> None:
    """Wrapper for entry point script."""
    cli = HistoryCli("freva")
    cli.parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    args = cli.parse_args(argv or sys.argv[1:])
    try:
        cli.run_cmd(args, **cli.kwargs)
    except KeyboardInterrupt:  # pragma: no cover
        print("KeyboardInterrupt, exiting", file=sys.stderr, flush=True)
        sys.exit(130)



"""Submit jobs using the moab workload manager."""
from __future__ import annotations
from .pbs import PBSJob
from typing import ClassVar


class MoabJob(PBSJob):
    submit_command: ClassVar[str] = "msub"
    cancel_command: ClassVar[str] = "canceljob"
    config_name: ClassVar[str] = "moab"



from django.db import models
from django.contrib.auth.models import User


class ToolPullRequest(models.Model):
    """
    Keep track of pull requests for tools
    """

    created = models.DateTimeField(auto_now_add=True)
    tool = models.CharField(max_length=50)
    tagged_version = models.CharField(max_length=50)
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    STATUS = [
        ("waiting", "waiting"),
        ("processing", "processing"),
        ("success", "success"),
        ("failed", "failed"),
    ]
    status = models.CharField(max_length=10, choices=STATUS)

    class Meta:
        app_label = "plugins"


class Version(models.Model):
    """
    The class belongs to a table holding all software versions used
    """

    #: Date and time when the process were scheduled
    timestamp = models.DateTimeField()
    #: Name of the tool
    tool = models.CharField(max_length=50)
    #: Version of the tool
    version = models.CharField(max_length=20)
    #: The tools internal version of a code versioning system
    internal_version_tool = models.CharField(max_length=40)
    #: The evaluation system's internal version
    internal_version_api = models.CharField(max_length=40)
    #: the repository to checkout thing
    repository = models.TextField()

    class Meta:
        app_label = "plugins"


class Parameter(models.Model):
    """
    Model for the tool parameter

    The entries tool and version seem to be redundant,
    but it could be necessary for not versioned tools.
    """

    class Impact(object):
        affects_values = 0
        affects_plots = 5
        no_effects = 9

    class Meta:
        app_label = "plugins"

    #: name of the parameter
    parameter_name = models.CharField(max_length=50)
    #: type of the parameter
    parameter_type = models.CharField(max_length=50)
    #: Name of the tool
    tool = models.CharField(max_length=50)
    #: Version of the tool
    version = models.IntegerField()
    #: mandatory
    mandatory = models.BooleanField()
    #: default value
    default = models.CharField(max_length=255, null=True, blank=True)
    #: how strong affects this parameter the output?
    IMPACT_CHOICES = (
        (Impact.affects_values, "Parameter affects values"),
        (Impact.affects_plots, "Parameter affects plots"),
        (Impact.no_effects, "No effects on output"),
    )

    impact = models.IntegerField(choices=IMPACT_CHOICES, default=Impact.affects_values)

    def __str__(self):
        return self.parameter_name



"""
Created on 31.05.2016

@author: Sebastian Illing
"""
import os
from pathlib import Path
import pytest
import shutil


def test_solr_search(dummy_solr):
    from evaluation_system.model.file import DRSFile

    # test path_only search
    res = dummy_solr.DRSFile.solr_search(path_only=True, variable="tauu")
    assert list(res) == [str(Path(dummy_solr.tmpdir) / dummy_solr.files[1])]

    # test drs search
    res = dummy_solr.DRSFile.solr_search(variable="ua")
    for i in res:
        assert isinstance(i, DRSFile)

    # use drs_structure
    res = dummy_solr.DRSFile.solr_search(drs_structure="cmip5")
    for j, i in enumerate(res):
        assert isinstance(i, DRSFile)
    assert j + 1 == 3


def test_compare(dummy_solr):
    fn2 = os.path.join(dummy_solr.tmpdir, dummy_solr.files[1])
    drs2 = dummy_solr.DRSFile.from_path(fn2)

    assert dummy_solr.drs == dummy_solr.drs
    assert not (dummy_solr.drs == drs2)
    assert not (drs2 == fn2)


def test_json_path(dummy_solr):
    j = dummy_solr.drs.to_json()
    assert isinstance(j, str)
    path = dummy_solr.drs.to_path()
    assert path == dummy_solr.fn


def test_find_structure_in_path(dummy_solr):

    search_path = str(Path(dummy_solr.tmpdir) / "cmip5")
    s = dummy_solr.DRSFile.find_structure_in_path(search_path)
    assert s == "cmip5"
    s = dummy_solr.DRSFile.find_structure_in_path(search_path, allow_multiples=True)
    assert s == ["cmip5"]
    with pytest.raises(ValueError):
        dummy_solr.DRSFile.find_structure_in_path("/no/valid/path")


def test_structure_from_path(dummy_solr):

    s = dummy_solr.DRSFile.find_structure_from_path(dummy_solr.fn)
    assert s == ["cmip5"]
    s = dummy_solr.DRSFile.find_structure_from_path(dummy_solr.fn, allow_multiples=True)
    assert s == ["cmip5"]
    with pytest.raises(ValueError):
        dummy_solr.DRSFile.find_structure_from_path("/no/valid/file_path")


def test_from_dict(dummy_solr):
    d = dummy_solr.drs.dict
    t = dummy_solr.DRSFile.from_dict(d, "cmip5")
    assert isinstance(t, dummy_solr.DRSFile)
    assert dummy_solr.drs.to_path() == t.to_path()


def test_from_json(dummy_solr):
    j = dummy_solr.drs.to_json()
    t = dummy_solr.DRSFile.from_json(j, "cmip5")
    assert isinstance(t, dummy_solr.DRSFile)
    assert dummy_solr.drs.to_path() == t.to_path()


def test_to_dataset(dummy_solr, dummy_crawl):
    from evaluation_system.model.file import DRSFile

    res = Path(dummy_solr.drs.to_dataset_path(versioned=True))
    root_dir = Path(dummy_solr.drs.dict["root_dir"])
    file = root_dir / Path(dummy_solr.files[0]).parent
    assert file == res
    drs = DRSFile.from_path(dummy_crawl[0])
    res = Path(drs.to_dataset_path(versioned=False))
    assert Path(dummy_crawl[0]).parent == res
    with pytest.raises(ValueError):
        drs.to_dataset_path(versioned=True)



"""
This class represents the foundation of all University of Cologne plugins. 

@author: Robert Redl
rredl@meteo.uni-koeln.de 
"""

from evaluation_system.model.file import *
from evaluation_system.model.solr import SolrFindFiles
import time
from datetime import datetime
import os
from Logger import Logger
from ShellScriptHelper import ShellScript
from NCLhelper import NCLscript
import copy
import re
import netCDF4


def print_arguments(config_dict):
    Logger.Indent("-> started with parameters:", 4, 7)
    params = []
    for key, value in config_dict.items():
        params.append("-> %-15s = %s" % (key, value))
    params.sort()
    for param in params:
        Logger.Indent(param, 8, 29)


def apply_workarounds_for_path(filelist):
    """
    some decadal experiments are not located in the same location then other experiments that belong together.
    the result ist that
    the files are grouped together in unexpected ways. here the path are replaced by predefined aliases.
    """
    # check if all file have the same root
    roots = []
    for onefile in filelist:
        if not onefile.dict["root_dir"] in roots:
            roots.append(onefile.dict["root_dir"])
    has_multiple_roots = len(roots) > 1

    # check all files
    for i in range(len(filelist)):
        # some baseline0 model runs are located in a different folder
        # this only matters if the root of the files is not always the same
        if has_multiple_roots:
            match = re.match(
                "/miklip/integration/data4miklip/projectdata/baseline0/output1/MPI-M/MPI-ESM-LR/decadal(\d\d\d\d)/(day|mon|6hr)/atmos/"
                + filelist[i].dict["parts"]["variable"]
                + "/"
                + filelist[i].dict["parts"]["ensemble"],
                filelist[i].to_path(),
            )
            if match is not None:
                newfile = splittedFile([filelist[i]])
                replacement = (
                    "/miklip/integration/data4miklip/model/global/miklip/baseline0/output1/MPI-M/MPI-ESM-LR/decadal%s/%s/atmos/%s/%s/v20111122/%s"
                    % (
                        match.group(1),
                        match.group(2),
                        filelist[i].dict["parts"]["cmor_table"],
                        filelist[i].dict["parts"]["ensemble"],
                        filelist[i].dict["parts"]["variable"],
                    )
                )
                newpath = filelist[i].to_path().replace(match.group(0), replacement)
                newfile.set_fake_path(newpath)
                Logger.Indent(
                    "-> workaround: %s => %s" % (filelist[i].to_path(), newpath), 8, 11
                )
                filelist[i] = newfile


def get_time_format_and_start_and_end(name_or_file, allow_no_time=False):
    """
    extracts the time format and and parts dictionary entry from a DRSFile or from a file name
    """
    # is it a file object?
    if isinstance(name_or_file, splittedFile) or isinstance(name_or_file, DRSFile):
        timepart = name_or_file.dict["parts"]["time"]
        # is it a valid time part?
        if re.match("\d{6,12}-\d{6,12}", timepart) is None:
            timepart_match = re.search(
                "\d{6,12}-\d{6,12}", name_or_file.dict["parts"]["file_name"]
            )
            if timepart_match is not None:
                timepart = timepart_match.group(0)
                # repair time part in dictionary
                name_or_file.dict["parts"]["time"] = timepart
            else:
                if not allow_no_time:
                    Logger.Error(
                        "Unable to extract time of file %s" % dfile.to_path(), -1
                    )
                else:
                    timepart = None
    # it is not a file object, that means it is a string containing a file name
    else:
        timepart_match = re.search("\d{6,12}-\d{6,12}", name_or_file)
        if timepart_match != None:
            timepart = timepart_match.group(0)
        else:
            if not allow_no_time:
                Logger.Error("Unable to extract time from string '%s'" % timestr, -1)
            else:
                timepart = None

    # check if we have a time part
    if timepart is None:
        return None, None, None
    else:
        # find the format of the date time string
        if len(timepart) == 17:
            datefmt = "%Y%m%d"
        elif len(timepart) == 21:
            datefmt = "%Y%m%d%H"
        elif len(timepart) == 23:
            datefmt = "%Y%m%d%H%M"
        elif len(timepart) == 13:
            datefmt = "%Y%m"
        else:
            Logger.Error("unable to get time_frequency from file '%s'" % name_or_file)
            Logger.Error(
                "only the time_frequencies 'day', '6hr', and 'mon' are so far supported!",
                -1,
            )
        # split and convert
        timepart_split = timepart.split("-")
        try:
            starttime = datetime.strptime(timepart_split[0], datefmt)
            endtime = datetime.strptime(timepart_split[1], datefmt)
        except ValueError as ve:
            Logger.Error(
                "The filename %s contains an invalid date: %s"
                % (dfile.to_path(), ve.message),
                -1,
            )
        return (datefmt, starttime, endtime)


def get_start_and_end_time_from_DRSFile(dfile, include_str=True, only_str=False):
    """
    returns a tuple (start,end,startstr,endstr)
    start,end       = date objects (only_str=False)
    startstr,endstr = string parts of the file name (only with include_str=True)
    """
    timepart = dfile.dict["parts"]["time"]
    # workaround for TRMM:
    if re.match("\d{6,12}-\d{6,12}", timepart) is None:
        timepart_match = re.search(
            "\d{6,12}-\d{6,12}", dfile.dict["parts"]["file_name"]
        )
        if timepart_match is not None:
            timepart = timepart_match.group(0)
            dfile.dict["parts"]["time"] = timepart
        else:
            Logger.Error("Unable to extract time of file %s" % dfile.to_path(), -1)
    # find the format of the date time string
    if len(timepart) == 17:
        datefmt = "%Y%m%d"
    elif len(timepart) == 21:
        datefmt = "%Y%m%d%H"
    elif len(timepart) == 25:
        datefmt = "%Y%m%d%H%M"
    elif len(timepart) == 13:
        datefmt = "%Y%m"
    else:
        Logger.Error("unable to get time_frequency from file '%s'" % dfile)
        Logger.Error(
            "only the time_frequencies 'day', '6hr', and 'mon' are so far supported!",
            -1,
        )
    # split and convert
    timepart_split = timepart.split("-")
    try:
        starttime = datetime.strptime(timepart_split[0], datefmt)
        endtime = datetime.strptime(timepart_split[1], datefmt)
    except ValueError as ve:
        Logger.Error(
            "The filename %s contains an invalid date: %s"
            % (dfile.to_path(), ve.message),
            -1,
        )
    if include_str and not only_str:
        return (starttime, endtime, timepart_split[0], timepart_split[1])
    elif include_str == True and only_str == True:
        return timepart_split[0], timepart_split[1]
    else:
        return starttime, endtime


def get_start_and_end_time_from_string(timestr, allow_no_time=False):
    timepart_match = re.search("\d{6,12}-\d{6,12}", timestr)
    if timepart_match is not None:
        timepart = timepart_match.group(0)
    else:
        if allow_no_time:
            return None, None
        else:
            Logger.Error("Unable to extract time from string '%s'" % timestr, -1)
    # find the format of the date time string
    if len(timepart) == 17:
        datefmt = "%Y%m%d"
    elif len(timepart) == 21:
        datefmt = "%Y%m%d%H"
    elif len(timepart) == 25:
        datefmt = "%Y%m%d%H%M"
    elif len(timepart) == 13:
        datefmt = "%Y%m"
    else:
        if allow_no_time:
            return None, None
        else:
            Logger.Error("unable to get time_frequency from time string '%s'" % timestr)
            Logger.Error(
                "only the time_frequencies 'day', '6hr', and 'mon' are so far supoorted!",
                -1,
            )
    # split and convert
    timepart_split = timepart.split("-")

    try:
        starttime = datetime.strptime(timepart_split[0], datefmt)
        endtime = datetime.strptime(timepart_split[1], datefmt)
    except ValueError as ve:
        if allow_no_time:
            return None, None
        else:
            Logger.Error(
                "The string '%s' contains an invalid date: %s" % (timestr, ve.message),
                -1,
            )
    return starttime, endtime


def solr_search_multivar(variables, ssargs):
    """
    use solr_search to find files for multiple variables
    """
    allfiles = []
    for variable in variables:
        ssargs["variable"] = variable
        allfiles.extend(DRSFile.solr_search(**ssargs))
    del ssargs["variable"]
    result = []
    for onefile in allfiles:
        if (
            os.path.exists(onefile.to_path())
            and not onefile.to_path().endswith(".nc.save")
            and not os.path.getsize(onefile.to_path()) == 0
        ):
            result.append(onefile)
        else:
            Logger.Warning(
                "the file '%s' does not exist, is empty, or ends with '.nc.save'!"
                % onefile.to_path()
            )
    return result


def number_of_files_in_folder(folder):
    """
    checks the presents of a folder and counts the number of files it contains.
    returns number of files found or -1 if the folder is not present.
    """
    if not os.path.exists(folder):
        return -1
    nfiles = 0
    for root, dirs, files in os.walk(folder):
        nfiles += len(files)
    return nfiles


class splittedFile(object):
    """
    Files of the same experiment are sometimes splitted up if they would become to large otherwise. These
    splitted files
    are can by grouped together in this class and treated as one.
    """

    def __init__(self, filelist, lead_year=None):
        """
        create a new splittedFile object from a list of DRSFile objects
        """
        self.parts = copy.deepcopy(filelist)
        self.dict = copy.deepcopy(filelist[0].dict)

        # get the first start time and the last end time
        first = None
        last = None
        for onefile in self.parts:
            # get the current start and end time of this file
            datefmt, start, end = get_time_format_and_start_and_end(onefile)
            if lead_year is not None:
                # calculate the new start and end time
                start = datetime(
                    year=start.year + lead_year - 1,
                    month=start.month,
                    day=start.day,
                    hour=start.hour,
                )
                end = datetime(
                    year=start.year, month=end.month, day=end.day, hour=end.hour
                )
            if first is None or start < first:
                first = start
            if last is None or end > last:
                last = end

        firsty = first.year
        lasty = last.year
        first = datetime(1900, first.month, first.day, first.hour, first.minute)
        last = datetime(1900, last.month, last.day, last.hour, last.minute)
        self.dict["parts"]["time"] = "%s%s-%s%s" % (
            str(firsty),
            first.strftime(datefmt)[4:],
            str(lasty),
            last.strftime(datefmt)[4:],
        )
        # merge variables
        # create at first a list of all variables
        variables = []
        for onefile in self.parts:
            if onefile.dict["parts"]["variable"] not in variables:
                variables.append(onefile.dict["parts"]["variable"])
        if len(variables) > 1:
            self.dict["parts"]["variable"] = "_".join(variables)
        # allow the path to be overridden
        self.manual_set_path = None

    def __str__(self):
        """
        get a string representation
        """
        return self.to_path(fake=False, separator="\n")

    def has_multiple_parts(self):
        return len(self.parts) > 1

    def to_path(self, fake=False, as_list=False, separator=" "):
        """
        returns a space separated string with all paths if fake == False,
        otherwise a path is generated which would fit to a single file with the same content
        """

        def get_path_string(x, pfake=False):
            if isinstance(x, splittedFile):
                return x.to_path(fake=pfake, separator="\n")
            else:
                return x.to_path()

        if fake:
            if self.manual_set_path is not None:
                return self.manual_set_path
            # get the first start time and the last end time
            return (
                get_path_string(self.parts[0], pfake=True)
                .replace(
                    self.parts[0].dict["parts"]["time"], self.dict["parts"]["time"]
                )
                .replace(
                    self.parts[0].dict["parts"]["variable"],
                    self.dict["parts"]["variable"],
                )
            )
        else:
            if as_list:
                result = []
                for onefile in self.parts:
                    if not isinstance(onefile, splittedFile):
                        result.append(onefile.to_path())
                    else:
                        result.extend(onefile.to_path(fake=False, as_list=True))
                return result
            else:
                return separator.join(map(lambda x: get_path_string(x), self.parts))

    def set_fake_path(self, new_path):
        """
        override the real path with a fake
        """
        self.manual_set_path = new_path

    @classmethod
    def get_fake_path(cls, onefile):
        """
        if the argument is an instance of splittedFile, then the fake path is returned. Otherwise the real path.
        """
        if isinstance(onefile, splittedFile):
            return onefile.to_path(fake=True)
        else:
            return onefile.to_path()


class basePlugin(object):
    def __init__(
        self, output=None, project="baseline1", model="mpi-esm-lr", experiment=None
    ):
        # store some information for later usage
        if output is None:
            raise Exception("No output directory specified!")
        subdir = "%s_%s_%s" % (time.strftime("%Y%m%d-%H%M%S"), project, model)
        subdir = subdir.replace("_*", "")
        if experiment is not None and experiment != "*":
            subdir += "_%s" % experiment.replace("*", "")
        # use the subdirectory only if the output and plotdir are the default dirs
        self.outputfiles = []
        self.outputfiles_ensstat = []

    def search_files(
        self,
        decadals=None,
        project="baseline1",
        firstyear=None,
        lastyear=None,
        product="*",
        time_frequency="6hr",
        model="mpi-esm-lr",
        ensembles=["*"],
        experiment=None,
        variable="ta",
        institute="MPI-M",
        realm="atmos",
        driving_model=None,
        rcm_ensemble=None,
        domain=None,
        find_variables=False,
    ):
        """
        use solr_search to find the files needed by the plugin
        set find_variables=True to get a list of available variables instead of searching files
        """
        # convert some arguments

        #         if '*' in ensembles:
        #             Logger.Error('All ensembles not allowed\n'
        #                          'Please select one',-1)

        if ensembles != "\*":
            ensembles = ensembles.split(",")
        else:
            ensembles = [ensembles]
        if decadals is not None:
            years = list(map(int, decadals))
        else:
            years = None
        if firstyear is not None:
            firstyear = int(firstyear)
        if lastyear is not None:
            lastyear = int(lastyear)

        # use solr_search to get the input files --------------------------------------------------
        # construct a search string for experiments
        experiment_prefix = experiment
        if experiment_prefix is None or experiment_prefix == "*":
            if project.lower() == "baseline0":
                experiment_prefix = "decadal"
            elif project.lower() == "baseline1":
                experiment_prefix = "decs4e"
            elif project.lower() == "prototype":
                experiment_prefix = "dffs4e"
            elif project.lower() == "cmip5":
                experiment_prefix = "decadal"
            elif project.lower() == "historical":
                experiment_prefix = "historical"
            else:
                experiment_prefix = "*"
        # if not experiment_prefix.endswith("*"):
        #    experiment_prefix += "*"

        # compose solr_search arguments
        ssargs = {}
        ssargs["project"] = project
        ssargs["institute"] = institute
        ssargs["realm"] = realm
        if type(variable) == list:
            variables = variable
        else:
            variables = [variable]
        ssargs["time_frequency"] = time_frequency
        # are there products in this project?
        product_facets = SolrFindFiles.facets(facets=["product"], **ssargs)
        if len(product_facets["product"]) > 0:
            ssargs["product"] = product
        # are there models in this project?
        model_facets = SolrFindFiles.facets(facets=["model"], **ssargs)
        if len(model_facets["model"]) > 0:
            ssargs["model"] = model
        # are there multiple experiments?
        experiment_facets = SolrFindFiles.facets(facets=["experiment"], **ssargs)
        if len(experiment_facets["experiment"]) > 0:
            ssargs["experiment"] = experiment_prefix
        # additional parameters for regional models
        if rcm_ensemble is not None:
            rcm_ensemble_facets = SolrFindFiles.facets(
                facets=["rcm_ensemble"], **ssargs
            )
            if len(rcm_ensemble_facets["rcm_ensemble"]) > 0:
                ssargs["rcm_ensemble"] = rcm_ensemble
        if driving_model is not None:
            driving_model_facets = SolrFindFiles.facets(
                facets=["driving_model"], **ssargs
            )
            if len(driving_model_facets["driving_model"]) > 0:
                ssargs["driving_model"] = driving_model
        if domain is not None:
            domain_facets = SolrFindFiles.facets(facets=["domain"], **ssargs)
            if len(domain_facets["domain"]) > 0:
                ssargs["domain"] = domain

        # search variables instead of files?
        if find_variables:
            variable_facets = SolrFindFiles.facets(facets=["variable"], **ssargs)
            return variable_facets["variable"]

        # put all files into a list

        self.inputfiles = []
        if years is not None or firstyear is not None or lastyear is not None:
            self.inputfilesByDecade = {}
            # we have multiple experiments that contain the decade
            if (
                "experiment" in ssargs
                and project != "observations"
                and project != "reanalysis"
                and years is not None
            ):
                for year in years:
                    yearfiles = []
                    ssargs["experiment"] = "%s%d" % (experiment_prefix, year)
                    for ens in ensembles:
                        if ens != "*":
                            ssargs["ensemble"] = ens
                        for onefile in solr_search_multivar(variables, ssargs):
                            self.inputfiles.append(onefile)
                            yearfiles.append(onefile)
                    self.inputfilesByDecade[year] = yearfiles

            # we have only one experiment. fetch all files and filter them by year
            elif firstyear is not None or lastyear is not None:
                for ens in ensembles:
                    if ens != "*":
                        ssargs["ensemble"] = ens
                    for onefile in solr_search_multivar(variables, ssargs):
                        starttime, endtime = get_start_and_end_time_from_DRSFile(
                            onefile, include_str=False
                        )
                        if firstyear is None and any(
                            [
                                True
                                for e in range(starttime.year, endtime.year + 1, 1)
                                if e <= lastyear
                            ]
                        ):
                            self.inputfiles.append(onefile)
                        elif lastyear is None and any(
                            [
                                True
                                for e in range(starttime.year, endtime.year + 1, 1)
                                if e >= firstyear
                            ]
                        ):
                            self.inputfiles.append(onefile)
                        elif any(
                            [
                                True
                                for e in range(firstyear, lastyear + 1, 1)
                                if e in range(starttime.year, endtime.year + 1, 1)
                            ]
                        ):
                            self.inputfiles.append(onefile)

            else:
                for ens in ensembles:
                    if ens != "*":
                        ssargs["ensemble"] = ens
                    for onefile in solr_search_multivar(variables, ssargs):
                        starttime, endtime = get_start_and_end_time_from_DRSFile(
                            onefile, include_str=False
                        )
                        for year in years:
                            if starttime.year > year and endtime.year <= year + 10:
                                self.inputfiles.append(onefile)
                                if year not in self.inputfilesByDecade:
                                    self.inputfilesByDecade[year] = [onefile]
                                else:
                                    self.inputfilesByDecade[year].append(onefile)
                                break
                                # we want all files, not only those for a special decade
        else:
            self.inputfilesByDecade = None
            for ens in ensembles:
                if ens != "*":
                    ssargs["ensemble"] = ens
                for onefile in solr_search_multivar(variables, ssargs):
                    self.inputfiles.append(onefile)

        # nothing found? cancel!
        if len(self.inputfiles) == 0:
            Logger.Error(
                "No input files found!\n"
                "Data-Browser command:\t"
                "freva --databrowser project='%s' product='%s' institute='%s' model='%s' experiment='%s'"
                " time_frequency='%s' realm='%s' variable='%s'"
                % (
                    project,
                    product,
                    institute,
                    model,
                    experiment,
                    time_frequency,
                    realm,
                    variable,
                ),
                -1,
            )

        # changed the time part if only a single lead year is of interest or remove files that do not belong to the
        # requested lead year

        # check for overlapping time-periods within the same folder
        self.inputfiles = self.remove_overlapping_time_periods_from_file_list(
            self.inputfiles
        )

        # check if all ensemble members have the same number of files
        self.inputfiles = self.check_ensemble_completeness(self.inputfiles)

        # repair some known special cases
        apply_workarounds_for_path(self.inputfiles)

        # merge multiple variables
        merged_by_var = self.merge_multiple_variables(self.inputfiles)
        if len(merged_by_var) == 0:
            Logger.Error("no files found for different variables and same time steps!")

            # group splitted files
            # self.inputfiles_merged = self.split_file_list_by_timecontinously(merged_by_var, return_splittedFile=True)

            # store information about the ensemble members to which the inputfiles belong
            # self.rcm_ensemble_members = Set()
            # for onfile in self.inputfiles:
            #    if "rcm_ensemble" in onfile.dict["parts"] and not onefile.dict["parts"]["rcm_ensemble"] in self.rcm_ensemble_members:
            #        self.rcm_ensemble_members.add(onefile.dict["parts"]["rcm_ensemble"])

    def group_input_by(self, files, groupby="ensemble", groupby2=None):
        """
        group files that belong to the same dataset and only differ by the ensemble member together.
        @param  files   list of DRSFile objects
        @returns        list of lists that contain ensemble groups, None if no group was found
        """

        def get_dataset_without_groupby(dfile):
            if isinstance(dfile, splittedFile):
                ds = os.path.basename(dfile.to_path(fake=True))
            else:
                ds = os.path.basename(dfile.to_path())
            dsens = dfile.dict["parts"][groupby]
            if groupby2 is not None and groupby2 in dfile.dict["parts"]:
                dsens2 = dfile.dict["parts"][groupby2]
                ds = ds.replace(dsens2, "<%s>" % groupby2)
            if groupby == "experiment":
                return dsens
            else:
                return ds.replace(dsens, "<%s>" % groupby)

        # check if it is possible to group the files by the given argument
        if groupby not in files[0].dict["parts"]:
            return None

        # put all file in a dictionary where the key is the dataset.
        # file that belong to the same dataset are placed in lists
        datasets = {}
        for onefile in files:
            dsname = get_dataset_without_groupby(onefile)
            if dsname in datasets:
                datasets[dsname].append(onefile)
            else:
                datasets[dsname] = [onefile]

        # remove groups that only contain one file
        keys_to_remove = []
        for key, value in datasets.items():
            if len(value) < 2:
                keys_to_remove.append(key)
        if len(keys_to_remove) > 0:
            for key in keys_to_remove:
                del datasets[key]

        # check time ranges for experiments
        keys_to_remove = []
        new_datasets = {}
        if groupby == "experiment":
            for key, value in datasets.items():
                newlist = self.split_file_list_by_timecontinously(value)
                if newlist is not None:
                    keys_to_remove.append(key)
                    for i in range(0, len(newlist)):
                        new_datasets["%s_%d" % (key, i)] = newlist[i]
            if len(keys_to_remove) > 0:
                for key in keys_to_remove:
                    del datasets[key]
                for key, value in new_datasets.items():
                    datasets[key] = value
            # remove again groups that only contain one file
            keys_to_remove = []
            for key, value in datasets.items():
                if len(value) < 2:
                    keys_to_remove.append(key)
            if len(keys_to_remove) > 0:
                for key in keys_to_remove:
                    del datasets[key]

        if len(datasets) > 0:
            return datasets
        else:
            return None

    def check_ensemble_completeness(self, files):
        """
        count the files for all ensemble member. remove members that have less members then others
        """
        # do nothing if we don't have an ensemble
        if "ensemble" not in files[0].dict["parts"]:
            return files
        # first step: group all files by their ensemble membership
        ensembles = {}
        for onefile in files:
            ensemble_part = onefile.dict["parts"]["ensemble"]
            if "rcm_ensemble" in onefile.dict["parts"]:
                ensemble_part += "-" + onefile.dict["parts"]["rcm_ensemble"]
            if ensemble_part not in ensembles:
                ensembles[ensemble_part] = [onefile]
            else:
                ensembles[ensemble_part].append(onefile)
        # find the maximum number of files per member
        max_files = 0
        for ensemble_part, filelist in ensembles.items():
            if len(filelist) > max_files:
                max_files = len(filelist)
        # copy all ensemble members with the maximal number of files to the result list
        result = []
        for ensemble_part, filelist in ensembles.items():
            if len(filelist) == max_files:
                result.extend(filelist)
            else:
                Logger.Indent(
                    "-> removed ensemble member %s, not enough files found!"
                    % ensemble_part,
                    8,
                    11,
                )
        # sort the result list
        return sorted(result, key=splittedFile.get_fake_path)

    def remove_overlapping_time_periods_from_file_list(self, filelist):
        """
        this is a workaround only. Removed are files that are located in the same folder but have overlapping
        time periods
        """
        # the list has to be sorted
        files = sorted(filelist, key=lambda x: x.to_path())
        # a dictionary for the periods, the folder is the key
        periods_per_folder = {}
        result = []
        for onefile in files:
            start, end = get_start_and_end_time_from_DRSFile(onefile, include_str=False)
            folder = os.path.dirname(onefile.to_path())
            if folder in periods_per_folder:
                old_period = periods_per_folder[folder]
                if old_period[0] <= start and old_period[1] >= start:
                    Logger.Warning(
                        "found overlapping time periods in folder '%s'" % folder
                    )
                    Logger.Warning("removing file '%s' from list" % onefile.to_path())
                else:
                    new_period = (old_period[0], end)
                    periods_per_folder[folder] = new_period
                    result.append(onefile)
            else:
                periods_per_folder[folder] = (start, end)
                result.append(onefile)
        return result

    def merge_multiple_variables(self, files):
        """
        creates a list of splitted files if more than one variable is included in the list of files.
        otherwise the list is returned unchanged.
        """
        # count the different variables
        variables = []
        for onefile in files:
            if onefile.dict["parts"]["variable"] not in variables:
                variables.append(onefile.dict["parts"]["variable"])
        variables.sort()
        # separate the files by variable
        files_by_var = []
        all_times = []
        all_ensemble = []
        for var in variables:
            file_list = {}
            for onefile in files:
                if onefile.dict["parts"]["variable"] == var:
                    time_part = onefile.dict["parts"]["time"]
                    ensemble_part = ""
                    if "ensemble" in onefile.dict["parts"]:
                        ensemble_part = onefile.dict["parts"]["ensemble"]
                    if "rcm_ensemble" in onefile.dict["parts"]:
                        ensemble_part += "-" + onefile.dict["parts"]["rcm_ensemble"]
                    file_list[time_part + ensemble_part] = onefile
                    if time_part not in all_times:
                        all_times.append(time_part)
                    if ensemble_part not in all_ensemble:
                        all_ensemble.append(ensemble_part)
            files_by_var.append(file_list)
        # create merged files from all time steps
        new_files = []
        for time_part in all_times:
            for ensemble_part in all_ensemble:
                file_list = []
                for ivar in range(len(variables)):
                    key = time_part + ensemble_part
                    if key in files_by_var[ivar]:
                        file_list.append(files_by_var[ivar][key])
                if len(file_list) == len(variables):
                    if len(file_list) == 1:
                        new_files.append(file_list[0])
                    else:
                        new_files.append(splittedFile(file_list))
                else:
                    Logger.Indent(
                        "-> %s not available for all variables in %s"
                        % (time_part, ensemble_part),
                        8,
                        11,
                    )
                    for onefile in file_list:
                        Logger.Indent("-> found only: %s" % onefile.to_path(), 12, 27)
        return sorted(new_files, key=splittedFile.get_fake_path)

    def split_file_list_by_timecontinously(self, files, return_splittedFile=False):
        """
        splits a list of files into a list of lists in case of time gaps
        @returns newlist or None if nothing was splitted.
        """

        def get_sort_key(onefile):
            """
            create a key for sorting the files depending on whether we have a selected lead_year or not
            """
            if self.lead_year is None:
                return onefile.to_path()
            else:
                key = ""
                if "ensemble" in onefile.dict["parts"]:
                    key += onefile.dict["parts"]["ensemble"]
                if "rcm_ensemble" in onefile.dict["parts"]:
                    key += onefile.dict["parts"]["rcm_ensemble"]
                key += os.path.basename(onefile.to_path())
                return key

        def get_dir_without_experiment(onefile):
            """
            if we use the lead_year, then files that belong together are not located in the same folder.
            remove the experiment in this case
            """
            if isinstance(onefile, splittedFile):
                pathstr = onefile.to_path(fake=True)
            else:
                pathstr = onefile.to_path()
            if self.lead_year is None:
                return os.path.dirname(pathstr)
            else:
                return os.path.dirname(pathstr).replace(
                    onefile.dict["parts"]["experiment"], ""
                )

        # sort the files in the way they likely extend each other
        sortedfiles = sorted(files, key=lambda x: get_sort_key(x))
        newlist = []
        newpart = []
        laststart = None
        lastend = None
        lastdir = get_dir_without_experiment(files[0])
        time_frequency = files[0].dict["parts"]["time_frequency"]
        # a gap of two days is allowed, because some models work with 30 day month
        if time_frequency == "day":
            max_delta_days = 2
        elif time_frequency == "6hr":
            max_delta_days = 2
        elif time_frequency == "mon":
            max_delta_days = 31
        else:
            Logger.Error(
                "unable to get time_frequency from file '%s'" % files[0].to_path()
            )
            Logger.Error(
                "only the time_frequencies 'day', '6hr', and 'mon' are so far supported!",
                -1,
            )
        for onefile in sortedfiles:
            starttime, endtime = get_start_and_end_time_from_DRSFile(
                onefile, include_str=False
            )
            currentdir = get_dir_without_experiment(onefile)
            if len(newpart) == 0:
                newpart.append(onefile)
            else:
                delta = starttime - lastend
                if (
                    delta.days > max_delta_days
                    or delta.days < 0
                    or currentdir != lastdir
                ):
                    newlist.append(newpart)
                    newpart = [onefile]
                else:
                    newpart.append(onefile)
            laststart = starttime
            lastend = endtime
            lastdir = currentdir
        if len(newpart) > 0:
            newlist.append(newpart)
        if len(newlist) > 0:
            if not return_splittedFile:
                return newlist
            else:
                # create splittedFile objects from the list
                sflist = list(map(lambda x: splittedFile(x), newlist))
                return sflist
        else:
            return None

    def change_time_part_to_lead_year(self, filelist, lead_year):
        """
        create splittedFile instances of the files given in filelist with a reduced time part in the file name
        """
        if lead_year > 10 or lead_year < 1:
            Logger.Error("Only lead years between 1 and 10 are allowed!", -1)
        newlist = []
        for onefile in filelist:
            start, end = get_start_and_end_time_from_DRSFile(onefile, include_str=False)
            if end.year - start.year + 1 < lead_year:
                Logger.Warning(
                    "file to short for lead year %d: %s"
                    % (lead_year, onefile.to_path())
                )
            else:
                newlist.append(splittedFile([onefile], lead_year=lead_year))
        return newlist

    def filter_inputfiles_by_lead_year(self, filelist, lead_year):
        """
        files those are created by split_file_list_by_time-continously and those contain single files only one year long
        are the input to this function. returned is a list of DSRfile objects that have the correct lead year
        """
        result = []
        for onefile in filelist:
            # find start and end for the merged file
            start, end = get_start_and_end_time_from_DRSFile(onefile, include_str=False)
            lead_start = datetime(start.year + lead_year - 1, 1, 1, 0, 0, 0)
            lead_end = datetime(start.year + lead_year - 1, 12, 31, 23, 59, 59)
            # now loop over all parts of this file to find the ones that belong to the correct lead year
            for onepart in onefile.parts:
                fstart, fend = get_start_and_end_time_from_DRSFile(
                    onepart, include_str=False
                )
                if fstart >= lead_start and fend <= lead_end:
                    result.append(onepart)
        return result

    def get_output_from_input_name(self, name, files=None):
        """
        construct the name of an output file from an input file
        """
        if name is not None and files is None:
            output_name = self.output + "/" + os.path.basename(name)
        if name is None and files is not None:
            # find first and last date to construct a new file name
            first = None
            last = None
            for onefile in files:
                (
                    starttime,
                    endtime,
                    startstr,
                    endstr,
                ) = get_start_and_end_time_from_DRSFile(onefile)
                if first is None:
                    first = starttime
                if last is None:
                    last = endtime
                if starttime < first:
                    first = starttime
                if endtime > last:
                    last = endtime
            newname = (
                files[0]
                .to_path()
                .replace(files[0].dict["parts"]["time"], "%s-%s" % (startstr, endstr))
            )
            return self.get_output_from_input_name(newname)
        return output_name

    def run_list_of_commands(self, commands, dryrun):
        # run the commands
        if dryrun:
            Logger.Info("\nDryrun! Not executed commands: ", color="red")
            for cmd in commands:
                if cmd.workpath is None:
                    print(cmd.getCommand() + "\n")
                else:
                    print("cd %s ; %s\n" % (cmd.workpath, cmd.getCommand()))
        else:
            SLURM_NTASKS_PER_NODE = os.getenv("SLURM_NTASKS_PER_NODE")
            if SLURM_NTASKS_PER_NODE is not None:
                SLURM_NTASKS_PER_NODE = int(SLURM_NTASKS_PER_NODE)
                Logger.Info(
                    "Number of processes limited by SLURM to %d" % SLURM_NTASKS_PER_NODE
                )
            exitcodes = ShellScript.run_scripts_parallel(
                commands, nproc=SLURM_NTASKS_PER_NODE
            )
            # check for any errors
            for c in exitcodes:
                if c[0] != 0:
                    raise Exception(
                        "exitcode: %d during parallel execution:\ncommand:\n%s\n\noutput:\n%s!"
                        % (c[0], c[2], c[1])
                    )

    def calculate_ensemble_stat(
        self, plugin_path, dryrun=False, separate=True, has_mon=False
    ):
        """
        calculate ensemble mean, min, max, etc.
        @separate   set to True if you don't want to add the files to the list of output files. The additional list
        outputfiles_ensstat only will be filled.
        """
        if self.ensemble_groups is None:
            Logger.Error("no ensemble groups!", -1)

        stats = ["ensmean", "ensmin", "ensmax"]

        # iterate over all groups and use cdo to calculate the statistics with CDO
        commands = []
        for key, value in self.ensemble_groups.items():
            group_files = []
            for stat in stats:
                output_name = self.get_output_from_input_name(
                    key.replace("<ensemble>", stat).replace("<rcm_ensemble>", stat)
                )
                group_files.append(output_name)
                if not separate:
                    self.outputfiles.append(output_name)
                command = ShellScript("cdo", check=False)
                command.addFlag("-O")
                command.addPositionalArgument(stat)
                for ifile in value:
                    if isinstance(ifile, splittedFile):
                        command.addPositionalArgument(
                            self.get_output_from_input_name(ifile.to_path(fake=True))
                        )
                    else:
                        command.addPositionalArgument(
                            self.get_output_from_input_name(ifile.to_path())
                        )
                command.addPositionalArgument(output_name)
                commands.append(command)
                # is there an additional output with the same names but a different extension
                if has_mon:
                    output_name2 = output_name.replace(".nc", "_mon.nc")
                    if not separate:
                        self.outputfiles.append(output_name2)
                    group_files.append(output_name2)
                    command2 = ShellScript("cdo", check=False)
                    command2.addPositionalArgument(stat)
                    for ifile in value:
                        if isinstance(ifile, splittedFile):
                            command2.addPositionalArgument(
                                self.get_output_from_input_name(
                                    ifile.to_path(fake=True)
                                ).replace(".nc", "_mon.nc")
                            )
                        else:
                            command2.addPositionalArgument(
                                self.get_output_from_input_name(
                                    ifile.to_path()
                                ).replace(".nc", "_mon.nc")
                            )
                    command2.addPositionalArgument(output_name2)
                    commands.append(command2)
            self.outputfiles_ensstat.append(group_files)

        # run the commands
        self.run_list_of_commands(commands, dryrun)

    def get_variable_unit(self, varname, filename):
        """
        open the netcdf file and read the unit attribute of the given variable (if available)
        """
        # check existence and open nc file
        if not os.path.exists(filename):
            Logger.Error("file not found: '%s' (get_variable_unit)" % filename, -1)
        nc = netCDF4.Dataset(filename, "r")

        # get the requested variable
        if varname not in nc.variables:
            Logger.Error(
                "variable '%s' not found in file '%s' (get_variable_unit)"
                % (varname, filename)
            )
        var = nc.variables[varname]
        units = var.units

        # close nc file
        nc.close()
        return units



"""Submit jobs to the lsf workload manager."""
from __future__ import annotations
from distutils.version import LooseVersion

import logging
import math
import os
from pathlib import Path
import re
import subprocess
import toolz
from typing import Any, cast, ClassVar, Optional, Coroutine, Union

from .core import Job

logger = logging.getLogger(__name__)


class LSFJob(Job):
    submit_command: ClassVar[str] = "bsub"
    cancel_command: ClassVar[str] = "bkill"
    config_name: ClassVar[str] = "lsf"

    def __init__(
        self,
        scheduler: Optional[str] = None,
        name: Optional[str] = None,
        queue: Optional[str] = None,
        project: Optional[str] = None,
        ncpus: Optional[int] = None,
        mem: Optional[int] = None,
        walltime: str = "",
        job_extra: Optional[list[str]] = None,
        lsf_units: Optional[str] = None,
        use_stdin: bool = False,
        **base_class_kwargs,
    ):
        super().__init__(scheduler=scheduler, name=name, **base_class_kwargs)

        self.use_stdin = use_stdin
        header_lines = []
        job_extra = job_extra or []
        # LSF header build
        if self.name is not None:
            header_lines.append("#BSUB -J %s" % self.job_name)
        if self.log_directory is not None:
            header_lines.append(
                "#BSUB -o %s/%s-%%J.out" % (self.log_directory, self.name or "worker")
            )
        if queue is not None:
            header_lines.append("#BSUB -q %s" % queue)
        if project is not None:
            header_lines.append('#BSUB -P "%s"' % project)
        if ncpus is None:
            # Compute default cores specifications
            ncpus = self.worker_cores
            logger.info(
                "ncpus specification for LSF not set, initializing it to %s" % ncpus
            )
        if ncpus is not None:
            header_lines.append("#BSUB -n %s" % ncpus)
            if ncpus > 1:
                # span[hosts=1] _might_ affect queue waiting
                # time, and is not required if ncpus==1
                header_lines.append('#BSUB -R "span[hosts=1]"')
        if mem is None:
            # Compute default memory specifications
            mem = cast(int, self.worker_memory)
            logger.info(
                "mem specification for LSF not set, initializing it to %s bytes" % mem
            )
        if mem is not None:
            lsf_units = lsf_units if lsf_units is not None else lsf_detect_units()
            memory_string = lsf_format_bytes_ceil(mem, lsf_units=lsf_units)
            header_lines.append("#BSUB -M %s" % memory_string)
        if walltime is not None:
            header_lines.append("#BSUB -W %s" % walltime)
        header_lines.extend(["#BSUB %s" % arg for arg in job_extra])

        # Declare class attribute that shall be overridden
        self.job_header = "\n".join(header_lines)

        logger.debug("Job script: \n %s" % self.job_script())

    def _submit_job(self, script_filename: Union[Path, str]) -> str:
        script_filename = str(script_filename)
        if self.use_stdin:
            piped_cmd = [self.submit_command + "< " + script_filename + " 2> /dev/null"]
            return self._call(piped_cmd, shell=True)
        else:
            return super()._submit_job(script_filename)


def lsf_format_bytes_ceil(n: int, lsf_units: str = "mb") -> str:
    """Format bytes as text

    Convert bytes to megabytes which LSF requires.

    Parameters
    ----------
    n: int
        Bytes
    lsf_units: str
        Units for the memory in 2 character shorthand, kb through eb

    Examples
    --------
    >>> lsf_format_bytes_ceil(1234567890)
    '1235'
    """
    lsf_units = lsf_units.lower()[0]
    converter = {"k": 1, "m": 2, "g": 3, "t": 4, "p": 5, "e": 6, "z": 7}
    return "%d" % math.ceil(n / (1000 ** converter[lsf_units]))


def lsf_detect_units() -> str:
    """Try to autodetect the unit scaling on an LSF system"""
    # Search for automatically, Using docs from LSF 9.1.3 for search/defaults
    unit = "kb"  # Default fallback unit
    try:
        # Start looking for the LSF conf file
        conf_dir = "/etc"  # Fall back directory
        # Search the two environment variables the docs say it could be at
        # (likely a typo in docs)
        for conf_env in ["LSF_ENVDIR", "LSF_CONFDIR"]:
            conf_search = os.environ.get(conf_env, None)
            if conf_search is not None:
                conf_dir = conf_search
                break
        conf_path = os.path.join(conf_dir, "lsf.conf")
        conf_file = open(conf_path, "r").readlines()
        # Reverse order search
        # (in case defined twice, get the one which will actually be processed)
        for line in conf_file[::-1]:
            # Look for very specific line
            line = line.strip()
            if not line.strip().startswith("LSF_UNIT_FOR_LIMITS"):
                continue
            # Found the line, infer the unit, only first 2 chars after "="
            unit = line.split("=")[1].lower()[0]
            break
        logger.debug(
            "Setting units to %s from the LSF config file at %s" % (unit, conf_file)
        )
    # Trap the lsf.conf does not exist, and the conf
    # file not setup right (i.e. "$VAR=xxx^" regex-form)
    except (EnvironmentError, IndexError):
        logger.debug(
            "Could not find LSF config or config file did not have "
            "LSF_UNIT_FOR_LIMITS set. Falling back to "
            "default unit of %s." % unit
        )
    return unit


@toolz.memoize
def lsf_version() -> Optional[LooseVersion]:
    out, _ = subprocess.Popen("lsid", stdout=subprocess.PIPE).communicate()
    versn: str = out.decode()
    match = re.search(r"(\d+\.)+\d+", versn)
    if match is not None:
        return LooseVersion(match.group())
    return None



"""Run job in the background on the same system."""
from __future__ import annotations
import logging
import os
from pathlib import Path
import subprocess
from typing import ClassVar, Optional

from .core import Job

logger = logging.getLogger(__name__)


class LocalJob(Job):
    __doc__ = """ Use Jobqueue with local bash commands

    This is mostly for testing.  It uses all the same machinery of
    jobqueue, but rather than submitting jobs to some external job
    queueing system, it launches them locally.

    Parameters
    ----------
    plugin args
    """

    config_name: ClassVar[str] = "local"
    cancel_command: ClassVar[str] = "kill -9"

    def __init__(
        self,
        name: Optional[str] = None,
        queue: Optional[str] = None,
        project: Optional[str] = None,
        resource_spec: Optional[str] = None,
        walltime: str = "",
        job_extra: Optional[list[str]] = None,
        env_extra: Optional[list[str]] = None,
        **kwargs,
    ):
        # Instantiate args and parameters from parent abstract class
        env_extra = env_extra or []
        env_extra = ["PID=$(pgrep -f $0)", "sleep 3"] + env_extra
        kwargs["memory"] = 1
        super().__init__(
            name=name, shebang="#!/usr/bin/env bash", env_extra=env_extra, **kwargs
        )

        # Declare class attribute that shall be overridden
        self.job_header = ""
        if self.log_directory:
            out_file = Path(self.log_directory) / f"{self.job_name}-$PID.local"
            self._command_template += f" &> {out_file}"
        logger.debug(f"Job script: \n {self.job_script()}")

    def _submit_job(self, script_filename):
        # Should we make this async friendly?
        cmd = ["/usr/bin/env", "bash", script_filename]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return str(process.pid)

    @classmethod
    def _close_job(cls, job_id, cancel_command):
        os.kill(int(job_id), 9)



"""
Created on 28.06.2016

@author: Sebastian Illing
"""
import logging
import mock
import os
from pathlib import Path
import pytest
import sys
import shutil

from git import Repo, Tag

from evaluation_system.tests import run_cli
from evaluation_system.tests import mockenv


def test_command_fail(capsys, admin_env, caplog):
    from evaluation_system.model.plugins.models import ToolPullRequest
    from django.contrib.auth.models import User

    # add an entry to pull-requests
    with mockenv(**admin_env):
        pr = ToolPullRequest.objects.create(
            tool="murcss",
            tagged_version="1.0",
            user=User.objects.first(),
            status="waiting",
        )
        # Clear all previous logs
        run_cli("check pull-request")
        # Get the last log on record
        _, loglevel, message = caplog.record_tuples[-1]
        assert loglevel == logging.ERROR
        assert message == "Plugin murcss does not exist"
        new_pr = ToolPullRequest.objects.get(id=pr.id)
        assert new_pr.status == "failed"


def test_command_success(capsys, dummy_git_path, git_config, admin_env):

    from evaluation_system.api import plugin_manager as pm
    from evaluation_system.model.plugins.models import ToolPullRequest
    from django.contrib.auth.models import User

    with mockenv(**admin_env):
        repo_path, tool_path = dummy_git_path
        this_dir = Path(__file__).parent
        shutil.copy(this_dir / "mocks" / "result_tags.py", repo_path)
        # prepare git repo
        os.system(
            (
                f"cd {repo_path}; git init; git config user.name tmp_user; "
                f"git config user.email test@testing.com; {git_config}"
            )
        )
        os.system('cd %s; git add *; git commit -m "first commit" ' % (repo_path))
        # clone it
        os.system("git clone %s %s" % (repo_path, tool_path))
        # create a new tag
        os.system('cd %s; git tag -a v2.0 -m "new tag"' % (repo_path))
        # add plugin to system
        with mock.patch.dict(
            os.environ, {"EVALUATION_SYSTEM_PLUGINS": f"{tool_path},result_tags"}
        ):
            pm.reload_plugins()
            repository = Repo(tool_path)
            assert len(repository.tags) == 0

            pr = ToolPullRequest.objects.create(
                tool="resulttagtest",
                tagged_version="v2.0",
                user=User.objects.first(),
                status="waiting",
            )
            # finally run the command
            run_cli("check pull-request")
            cmd_out = capsys.readouterr().out
            assert "Processing pull request for resulttagtest by" in cmd_out
            new_pr = ToolPullRequest.objects.get(id=pr.id)
            assert new_pr.status == "success"
            assert repository.tags[0].name == "v2.0"
            assert len(repository.tags) == 1


def test_get_version(admin_env):
    from evaluation_system.model.repository import get_version

    # self version test
    version = get_version(".")
    assert len(version) == 2

    not_versioned = get_version("/tmp")
    assert not_versioned == ("unknown", "unknown")


def test_forbidden_usage(dummy_env):
    from freva.cli.admin.check import check4pull_request

    with pytest.raises(RuntimeError):
        check4pull_request()

    with pytest.raises(SystemExit):
        with pytest.raises(RuntimeError):
            run_cli("check pull-request")



"""Definition of the base class that is used to implement user plugins.

The plugin API (Application Program Interface) is the central connection
of a user plugin code and the Freva system infrastructure. The API
enables the users to conveniently set up, run, and keep track of applied plugins.
The reference below gives an overview of how to set up user defined plugins.
For this purpose we assume that a plugin core (without Freva) already exists
- for example as a command line interface tool (cli). Once such a cli has been
set up a interface to Freva must be defined. This reference will introduce
the possible definition options below.

Here we assume that the above mentioned cli code is stored in a directory
name ``/mnt/freva/plugins/new_plugin`` furthermore the actual cli code can be
excecuted via:
    .. code-block:: console

        cli/calculate -c 5 -n 6.4 --overwrite --name=Test

With help of this API a Freva plugin can be created in ``/mnt/freva/plugin/new_plugin/plugin.py``

"""
from __future__ import annotations
import abc
from configparser import ConfigParser, ExtendedInterpolation
from contextlib import contextmanager
from datetime import datetime
from functools import partial
import logging
import os
from pathlib import Path
import re
import subprocess as sub
import sys
import stat
import shlex
import shutil
import socket
import tempfile
import traceback
import textwrap
from time import time
from typing import cast, Any, Dict, IO, Optional, Union, Iterator, Iterable, TextIO
from uuid import uuid4


from PyPDF2 import PdfReader

from evaluation_system.model.user import User
import evaluation_system.model.history.models as hist_model
import evaluation_system.model.repository as repository
from evaluation_system.misc.utils import TemplateDict
from evaluation_system.misc import config, logger as log
from evaluation_system.misc.exceptions import (
    ConfigurationException,
    deprecated_method,
    hide_exception,
)

from evaluation_system.misc.utils import PIPE_OUT
from evaluation_system.model.solr_core import SolrCore
from .workload_manager import schedule_job
from .user_data import DataReader

__version__ = (1, 0, 0)

ConfigDictType = Dict[str, Optional[Union[str, float, int, bool]]]


class PluginAbstract(abc.ABC):
    """Base class that is used as a template for all Freva plugins.
    Any api wrapper class defining Freva plugins must inherit from this class.

    Parameters
    ----------
    user : evaluation_system.model.user.User, default None
          Pre defined evaluation system user object, if None given (default)
          a new instance of a user object for the current user will be created.


    The following attributes are mandatory and have to be set:
        * :class:`__short_description__`
        * :class:`__version__`
        * :class:`__parameters__`
        * :class:`run_tool`

    Whereas the following properties can be optionally set:
        * :class:`__long_description__`
        * :class:`__tags__`
        * :class:`__category__`



    Minimal Example
    ---------------

    In order to configure and call this cli, a Freva wrapper api class will
    have the be created in ``/mnt/freva/plugins/new_plugin/plugin.py``.
    A minimal configuration example would look as follows:

    .. code-block:: python

        from evaluation_system.api import plugin, parameters
        class MyPlugin(plugin.PluginAbstract):
            __short_description__ = "Short plugin description"
            __long_description__ = "Optional longer description"
            __version__ = (2022, 1, 1)
            __parameters__ =  parameters.ParameterDictionary(
                                parameters.Integer(
                                    name="count",
                                    default=1,
                                    help=("This is an optional configurable "
                                          "int variable named number without "
                                          "default value and this description")
                                ),
                                parameters.Float(
                                    name="number",
                                    mandatory=True,
                                    help="Required float value without default"
                                ),
                                parameters.Bool(
                                    name="overwrite",
                                    default=False,
                                    help=("a boolean parameter "
                                          "with default value of false")
                                ),
                                parameters.String(name='str')
                              )
            def run_tool(
                self, config_dict: dict[str, str|int|bool]
            ) -> None:
                '''Definition of the tool the is running the cli.

                Parameters:
                -----------
                config_dict: dict
                    Plugin configuration stored in a dictionary
                '''

                self.call(
                    (
                      f"cli/calculate -c {config_dict['count']} "
                      f"-n {config_dict['number']} --name={config_dict['name']}
                    )
                )
                print("MyPlugin was run with", config_dict)

    .. note::
        The actual configuration is defined by the :class:`__parameters__`
        property, which is of type
        :class:`evaluation_system.api.parameters.ParameterDictionary`.


    If you need to test it use the ``EVALUATION_SYSTEM_PLUGINS`` environment
    variable to point to the source directory and package.
    For example assuming you have the source code in ``/mnt/freva/plugins``
    and the package holding the class implementing
    :class:`evaluation_system.api.plugin` is
    ``my_plugin.plugin`` (i.e. its absolute file path is
    ``/mnt/freva/plugins/my_plugin/plugin_module.py``), you would tell the system
    how to find the plugin by issuing the following command (bash & co)::

        export EVALUATION_SYSTEM_PLUGINS=/mnt/freva/plugins/my_plugin,plugin

    Use a colon to separate multiple items::

        export EVALUATION_SYSTEM_PLUGINS=/path1,plguin1:/path2,plugin2:/path3,plugin3

    By telling the system where to find the packages it can find the
    :class:`evaluation_system.api.plugin` implementations. The system just
    loads the packages and get to the classes using the
    :py:meth:`class.__subclasses__` method. The reference speaks about
    *weak references* so it's not clear if (and when) they get removed.
    We might have to change this in the future if it's not enough.
    Another approach would be forcing self-registration of a
    class in the :ref:`__metaclass__ <python:datamodel>` attribute when the
    class is implemented.

    """

    special_variables: Optional[dict[str, str]] = None
    """This dictionary resolves the *special variables* that are available

    to the plugins in a standardized manner. These variables are initialized per user and
    plugin. They go as follow:

    ================== ==================================================================
    Variables          Description
    ================== ==================================================================
    USER_BASE_DIR      Absolute path to the central directory for this user.
    USER_OUTPUT_DIR    Absolute path to where the plugin outputs for this user are stored.
    USER_PLOTS_DIR     Absolute path to where the plugin plots for this user are stored.
    USER_CACHE_DIR     Absolute path to the cached data (temp data) for this user.
    USER_UID           The users' User IDentifier
    SYSTEM_DATE        Current date in the form YYYYMMDD (e.g. 20120130).
    SYSTEM_DATETIME    Current date in the form YYYYMMDD_HHmmSS (e.g. 20120130_101123).
    SYSTEM_TIMESTAMP   Milliseconds since epoch (i.e. e.g. 1358929581838).
    SYSTEM_RANDOM_UUID A random Universal Unique Identifier (e.g. 912cca21-6364-4f46-9b03-4263410c9899).
    ================== ==================================================================

    A plugin/user might then use them to define a value in the following way::

        output_file='$USER_OUTPUT_DIR/myfile_${SYSTEM_DATETIME}blah.nc'

    """

    tool_developer: Optional[str] = None
    """Name of the developer who is responsible for the tool."""

    def __init__(self, *args, user: Optional[User] = None, **kwargs) -> None:
        """Plugin main constructor.

        It is designed to catch all calls. It accepts a ``user``
        argument containing an :class:`evaluation_system.model.user.User`
        representing the user for which this plugin will be created.
        It is used here for setting up the user-defined configuration but
        the implementing plugin will also have access to it. If no user
        is provided an object representing the current user, i.e. the user
        that started this program, is created.
        """
        self._user = user or User()
        # id of row in history table I think
        # this was being spontaneously created when running the plugin which
        # works for now because it creates a new instance on every run
        self.rowid = 0
        self._plugin_out: Optional[Path] = None

        # this construct fixes some values but allow others to be computed on
        # demand it holds the special variables that are accessible to both users
        # and developers
        # self._special_vars = SpecialVariables(self.__class__.__name__, self._user)

        plugin_name, user = self.__class__.__name__.lower(), self._user
        self._special_variables = TemplateDict(
            USER_BASE_DIR=user.getUserBaseDir,
            USER_CACHE_DIR=partial(
                user.getUserCacheDir, tool=plugin_name, create=False
            ),
            USER_PLOTS_DIR=partial(
                user.getUserPlotsDir, tool=plugin_name, create=False
            ),
            USER_OUTPUT_DIR=partial(
                user.getUserOutputDir, tool=plugin_name, create=False
            ),
            USER_UID=user.getName,
            SYSTEM_DATE=lambda: datetime.now().strftime("%Y%m%d"),
            SYSTEM_DATETIME=lambda: datetime.now().strftime("%Y%m%d_%H%M%S"),
            SYSTEM_TIMESTAMP=lambda: str(int(time() * 1000)),
            SYSTEM_RANDOM_UUID=lambda: str(uuid4()),
        )

    @property
    @abc.abstractmethod
    def __version__(self) -> tuple[int, int, int]:
        """3-value tuple representing the plugin version.

        Example
        -------

        >>>    verion = (1, 0, 3)
        """
        raise NotImplementedError("This attribute must be implemented")

    @property
    def __long_description__(self) -> str:
        """Optional long description of this plugin."""
        return ""

    @property
    @abc.abstractmethod
    def __short_description__(self) -> str:
        """Mandatory short description of this plugin.

        It will be displayed to the user in the help and
        when listing all plugins.
        """
        raise NotImplementedError("This attribute must be implemented")

    @property
    @abc.abstractmethod
    def __parameters__(self):
        """Mandatory definitions of all known configurable parameters."""
        raise NotImplementedError("This attribute must be implemented")

    @property
    def __category__(self) -> str:
        """Optional category this plugin belongs to."""
        return ""

    @property
    def __tags__(self) -> list[str]:
        """Optional tags, that are the plugin can be described with."""
        return [""]

    def run_tool(self, config_dict: Optional[ConfigDictType] = None) -> Optional[Any]:
        """Method executing the tool.

        The method should be overidden by the custom plugin tool method.

        Parameters
        ----------
        config_dict:
            A dict with the current configuration (param name, value)
            which the tool will be run with

        Returns
        -------
        list[str]:
            Return values of :class:`prepare_output([<list_of_created_files>])` method
        """
        raise NotImplementedError("This method must be implemented")

    def _execute(self, cmd: list[str], check=True, **kwargs):
        # Do not allow calling shell=True
        kwargs["shell"] = False
        kwargs["stdout"] = sub.PIPE
        kwargs["stdin"] = None
        kwargs["universal_newlines"] = True
        kwargs.setdefault("stderr", sub.STDOUT)
        with sub.Popen(cmd, **kwargs) as res:
            stdout = cast(IO[Any], res.stdout)
            for line in iter(stdout.readline, ""):
                print(line, end="", flush=True)
            return_code = res.wait()
            if return_code and check:
                log.error("An error occured calling %s", cmd)
                log.error("Check also %s", self.plugin_output_file)
                raise sub.CalledProcessError(
                    return_code,
                    cmd,
                )
            return res

    @property
    def plugin_output_file(self) -> Path:
        """Filename where stdout is written to."""
        if self._plugin_out is None:
            pid = os.getpid()
            plugin_name = self.__class__.__name__
            log_directory = os.path.join(
                self._user.getUserSchedulerOutputDir(),
                plugin_name.lower(),
            )
            self._plugin_out = Path(log_directory) / f"{plugin_name}-{pid}.local"
        return self._plugin_out

    def _set_interactive_job_as_running(self, rowid: Optional[int]):
        """Set an interactive job as running."""
        host = socket.getfqdn()
        try:
            hist = hist_model.History.objects.get(id=rowid)
            hist.slurm_output = str(self.plugin_output_file)
            hist.host = host.partition(".")[0]
            hist.status = hist_model.History.processStatus.running
            hist.save()
        except hist_model.History.DoesNotExist:
            pass

    @contextmanager
    def _set_environment(
        self, rowid: Optional[int], is_interactive_job: bool
    ) -> Iterator[None]:
        """Set the environement."""
        env_path = os.environ["PATH"]
        stdout = [sys.stdout]
        stderr = [sys.stderr]
        log_stream_handle: Optional[logging.StreamHandler] = None
        try:
            self.plugin_output_file.touch(mode=0o2755)
        except FileNotFoundError:
            self.plugin_output_file.parent.mkdir(parents=True, mode=0o2777)
            self.plugin_output_file.touch(mode=0o2755)
        try:
            os.environ["PATH"] = f"{self.conda_path}:{env_path}"
            if is_interactive_job is True:
                f = self.plugin_output_file.open("w")
                self._set_interactive_job_as_running(rowid)
                stdout.append(f)
                stderr.append(f)
            with PIPE_OUT(*stdout) as p_sto, PIPE_OUT(*stderr) as p_ste:
                sys.stdout = p_sto
                sys.stderr = p_ste
                log_stream_handle = logging.StreamHandler(p_ste)
                log.addHandler(log_stream_handle)
                yield
        except Exception as error:
            if is_interactive_job is True:
                traceback.print_exc(file=f)
            raise error
        finally:
            os.environ["PATH"] = env_path
            sys.stdout = stdout[0]
            sys.stderr = stderr[0]
            if is_interactive_job is True:
                if log_stream_handle is not None:
                    log.removeHandler(log_stream_handle)
                f.flush()
                f.close()

    @property
    def conda_path(self) -> str:
        """Add the conda env path of the plugin to the environment.i

        :meta private:
        """

        from evaluation_system.api import plugin_manager as pm

        plugin_name = self.__class__.__name__.lower()
        try:
            plugin_path = Path(pm.get_plugins()[plugin_name].plugin_module)
        except KeyError:
            return ""
        return f"{plugin_path.parent / 'plugin_env' / 'bin'}"

    def _run_tool(
        self,
        config_dict: Optional[ConfigDictType] = None,
        unique_output: bool = True,
        out_file: Optional[Path] = None,
        rowid: Optional[int] = None,
    ) -> Optional[Any]:
        config_dict = self._append_unique_id(config_dict, unique_output)
        if out_file is None:
            is_interactive_job = True
        else:
            is_interactive_job = False
            self._plugin_out = out_file
        for key in config.exclude:
            config_dict.pop(key, "")
        with self._set_environment(rowid, is_interactive_job):
            try:
                result = self.run_tool(config_dict=config_dict)
            except NotImplementedError:
                result = deprecated_method("PluginAbstract", "run_tool")(self.runTool)(config_dict=config_dict)  # type: ignore
            return result

    def _append_unique_id(
        self, config_dict: Optional[ConfigDictType], unique_output: bool
    ) -> ConfigDictType:
        from evaluation_system.api.parameters import Directory, CacheDirectory

        config_dict = config_dict or {}
        for key, param in self.__parameters__.items():
            tmp_param = self.__parameters__.get_parameter(key)
            if isinstance(tmp_param, (Directory, CacheDirectory)) and unique_output:
                if key in config_dict.keys() and config_dict[key] is not None:
                    config_dict[key] = os.path.join(
                        str(config_dict[key]), str(self.rowid)
                    )
        return config_dict

    def quitnkill(self):
        """If error occurs quit python and kill child processes by group ID

        :meta private:
        """
        PID = os.getpid()
        self.call(f'setsid nohup bash -c "kill -9 -- -{PID}"  </dev/null &>/dev/null &')
        raise SystemExit

    @deprecated_method("PluginAbstract", "add_output_to_databrowser")
    def linkmydata(self, *args, **kwargs):  # pragma: no cover
        """Deprecated version of the :class:`add_output_to_databrowser` method.

        :meta private:
        """
        return self.add_output_to_databrowser(**args, **kwargs)

    def add_output_to_databrowser(
        self,
        plugin_output: os.PathLike,
        *,
        model: str = "freva",
        institute: Optional[str] = None,
        ensemble: str = "r0i0p0",
        time_frequency: Optional[str] = None,
        variable: Optional[str] = None,
    ) -> Path:
        """Add Plugin output data to the solr database.

        This methods crawls the plugin output data directory and adds
        any files that were found to the apache solr database.

        ..note::
            Use the ``ensemble`` and ``model``
            arguments to specify the search facets that are going to be
            added to the solr server for this pluign run. This will help
            users better to better distinguish their plugin result search.

        The following facets are fixed:

        - project: ``<project_name>-plugin-ouput``
        - product: ``user-<user_name>``
        - experiment: ``<plugin_name>``
        - dataset version: ``<history_id>``
        - realm: ``plugins``
        - cmor_table: ``<plugin_version>``

        Parameters
        ----------
        plugin_output: os.PathLike
            Plugin output directory or file created by the files. If a
            directory is given all data files within the sub directories
            will be collected for ingestion.
        model: str, default: None
            Default model facet. If None is given (default) the model name will
            be set to ``freva``. The model argument can be used to distinguish
            plugin results for different setups.
        institute: str, default: None
            Default institute facet. Use the argument to make plugin results
            from various setups better distinguishable. If None given (default)
            the institute will be set the the freva project name.
        ensemble: str, default: r0i0p0
            Default ensemble facet. Like for model the ensemble argument should
            be used to distinguish plugins results with different setups.
        time_frequency: str, default: None
            Default time frequency facet. If None is given (default) the time
            frequency will be retrieved from the output files.
        variable: str, default: None
            Default variable facet. If None is given (default) the variable
            will be retrieved from the output files.

        Returns
        -------
            Path: Path to the new directory that containes the data.
        """
        _, plugin_version = repository.get_version(self.wrapper_file)
        plugin_version = plugin_version or "no_plugin_version"
        product_dir = f"{config.get('project_name')}-plugin-results"
        root_dir = DataReader.get_output_directory() / f"user-{self._user.getName()}"
        drs_config = dict(
            project=root_dir.name,
            product=product_dir,
            model=model,
            experiment=self.__class__.__name__.lower(),
            realm="plugins",
            institute=institute or root_dir.name,
            ensemble=ensemble,
            cmor_table=plugin_version,
            version=f"v{self.rowid}",
        )
        if time_frequency:
            drs_config["time_frequency"] = time_frequency
        if variable:
            drs_config["variable"] = variable
        user_data = DataReader(plugin_output, **drs_config)
        for output_file in user_data:
            new_file = user_data.file_name_from_metdata(output_file)
            new_file.parent.mkdir(exist_ok=True, parents=True, mode=509)
            shutil.copy(str(output_file), str(new_file))
        SolrCore.load_fs(root_dir / product_dir, drs_type=user_data.drs_specification)
        return root_dir / product_dir

    @deprecated_method("PluginAbstract", "prepare_output")
    def prepareOutput(self, *args) -> dict[str, dict[str, str]]:
        """Deprecated method for :class:`prepare_output`.

        :meta private:
        """
        return self.prepare_output(*args)

    def prepare_output(
        self, output_files: Union[str, list[str], dict[str, dict[str, str]]]
    ) -> dict[str, dict[str, str]]:
        """Prepare output for files supposedly created.


        This method checks if the files exist and returns a dictionary with
        information about them::

            { <absolute_path_to_file>: {
                'timestamp': os.path.getctime(<absolute_path_to_file>),
                'size': os.path.getsize(<absolute_path_to_file>)
                }
            }

        Use it for the return call of run_tool.

        Parameters
        ----------
        output_files:
            iterable of strings or single string with paths to all files that
            where created by the tool.

        Returns
        -------
        dict:
            dictionary with the paths to the files that were created as
            key and a dictionary as value.
        """
        result = {}
        metadata: dict[str, str] = {}
        if isinstance(output_files, str):
            output_files = [output_files]
        for file_path in output_files:
            # we expect a meta data dictionary
            if isinstance(output_files, dict):
                if not isinstance(output_files[file_path], dict):
                    raise ValueError("Meta information must be of type dict")
                metadata = output_files[file_path]
            if os.path.isfile(file_path):

                self._extend_output_metadata(file_path, metadata)
                result[os.path.abspath(file_path)] = metadata
            elif os.path.isdir(file_path):
                # ok, we got a directory, so parse the contents recursively
                for file_path in [
                    os.path.join(r, f)
                    for r, _, files in os.walk(file_path)
                    for f in files
                ]:
                    filemetadata = metadata.copy()
                    self._extend_output_metadata(file_path, filemetadata)

                    # update meta data with user entries
                    usermetadata = result.get(os.path.abspath(file_path), {})
                    filemetadata.update(usermetadata)

                    result[os.path.abspath(file_path)] = filemetadata
            else:
                result[os.path.abspath(file_path)] = metadata
        return result

    def _extend_output_metadata(self, file_path, metadata):
        fstat = os.stat(file_path)
        if "timestamp" not in metadata:
            metadata["timestamp"] = fstat[stat.ST_CTIME]
            # metadata['timestamp'] = os.path.getctime(file_path)
        if "size" not in metadata:
            metadata["size"] = fstat[stat.ST_SIZE]
            # metadata['size'] = os.path.getsize(file_path)
        if "type" not in metadata:
            ext = os.path.splitext(file_path)
            if ext:
                ext = ext[-1].lower()
                if ext in ".jpg .jpeg .png .gif .mp4 .mov".split():
                    metadata["type"] = "plot"
                    metadata["todo"] = "copy"

                if ext in ".tif .svg .ps .eps".split():
                    metadata["type"] = "plot"
                    metadata["todo"] = "convert"

                if ext == ".pdf":

                    # If pdfs have more than one page we don't convert them,
                    # instead we offer a download link
                    pdf = PdfReader(open(file_path, "rb"))
                    num_pages = len(pdf.pages)
                    metadata["type"] = "pdf"
                    if num_pages > 1:
                        metadata["todo"] = "copy"
                    else:
                        metadata["todo"] = "convert"

                if ext in ".tex".split():
                    metadata["type"] = "plot"

                elif ext in ".nc .bin .ascii".split():
                    metadata["type"] = "data"
                if ext in [".zip"]:
                    metadata["type"] = "pdf"
                    metadata["todo"] = "copy"
                elif ext in [".html", ".xhtml"]:
                    metadata["todo"] = "copy"

    @deprecated_method("PluginAbstract", "get_help")
    def getHelp(self, **kwargs) -> str:
        """Deprecated version of the :class:`get_help` method.

        :meta private:
        """
        return self.get_help(**kwargs)

    def get_help(self, width: int = 80) -> str:
        """Representation of the help string

        This method uses the information from the implementing class name,
        :class:`__version__`, :class:`__short_description__` and
        :class:`__config_metadict__` to create a proper help. Since it returns
        a string, the implementing class might use it and extend it if required.

        Parameters
        ----------
        width:
            Wrap text to this width.

        Returns
        -------
        str:
            A string containing the help.


        :meta private:
        """
        help_txt = self.__long_description__.strip()
        if not help_txt:
            help_txt = self.__short_description__
        return "{} (v{}): {}\n{}".format(
            self.__class__.__name__,
            ".".join([str(i) for i in self.__version__]),
            help_txt,
            self.__parameters__.get_help(),
        )

    def get_current_config(self, config_dict: Optional[ConfigDictType] = None) -> str:
        """Retreive the plugin configuration as string representation.

        Parameters
        ----------
            config_dict:
                the dict containing the current configuration being displayed.
                This info will update the default values.

        Returns
        -------
        str:
            The current configuration in a string for displaying.


        :meta private:
        """
        max_size = max([len(k) for k in self.__parameters__])
        config_dict = config_dict or {}
        current_conf = []
        config_dict_resolved = self.setup_configuration(
            config_dict=config_dict, check_cfg=False
        )
        config_dict_orig = cast(Dict[str, Any], dict(self.__parameters__))
        config_dict_orig.update(config_dict or {})

        def show_key(key: str) -> str:
            """Format the configuration values.
            This functions formats the configuration depending on whether the
            configuration values contain variables or not.

            Returns
            -------
            str

            :meta private:
            """
            if config_dict_resolved[key] == config_dict_orig[key]:
                return config_dict_orig[key]
            return f"{config_dict_orig[key]} [{config_dict_resolved[key]}]"

        for key in self.__parameters__:
            line_format = "%%%ss: %%s" % max_size

            if key in config_dict:
                # user defined
                curr_val = show_key(key)
            else:
                # default value
                default_value = self.__parameters__[key]
                if default_value is None:
                    if self.__parameters__.get_parameter(key).mandatory:
                        curr_val = "- *MUST BE DEFINED!*"
                    else:
                        curr_val = "-"
                else:
                    curr_val = "- (default: %s)" % show_key(key)

            current_conf.append(line_format % (key, curr_val))

        return "\n".join(current_conf)

    @deprecated_method("PluginAbstract", "class_basedir")
    def getClassBaseDir(self) -> Optional[str]:
        """Deprecated method for class_basedir.

        :meta private:
        """
        return self.class_basedir

    @property
    def class_basedir(self) -> str:
        """Get absolute path to the module defining the plugin class."""
        return os.path.join(
            *self._split_path(self.wrapper_file)[: -len(self.__module__.split("."))]
        )

    @property
    def wrapper_file(self) -> str:
        """Get the location of the wrapper file."""
        module_path: Optional[str] = sys.modules[self.__module__].__file__
        return os.path.abspath(module_path or "")

    @property
    def user(self) -> User:
        """Return the user class for which this instance was generated."""
        return self._user

    def parse_config_str_value(
        self, param_name: str, str_value: str, fail_on_missing: bool = True
    ) -> Optional[str]:
        """Parse the string in ``str_value`` into the most appropriate value.

        The string *None* will be mapped to the value ``None``.
        On the other hand the quoted word *"None"* will remain as the string
        ``"None"`` without any quotes.

        Parameters
        ----------
        param_name:
            Parameter name to which the string belongs.
        str_value:
            The string that will be parsed.
        fail_on_missing:
            If the an exception should be risen in case the param_name is not
            found in :class:`__parameters__`

        Returns
        -------
        str:
            the parsed string, or the string itself if it couldn't be parsed,
            but no exception was thrown.

        Raises
        ------
        ( :class:`ConfigurationException` ) if parsing couldn't succeed.

        :meta private:
        """

        if str_value == "None" or str_value == '"None"':
            return None

        if self.__parameters__ is None or (
            not fail_on_missing and param_name not in self.__parameters__
        ):
            # if there's no dictionary reference or the param_name is not in it
            # and we are not failing
            # just return the str_value
            return str_value

        else:
            return self.__parameters__.get_parameter(param_name).parse(str_value)

    @deprecated_method("PluginAbstract", "setup_configuration")
    def setupConfiguration(
        self, **kwargs
    ) -> dict[str, Union[str, int, float, bool, None]]:  # pragma: no cover
        """Deprecated version of the :class:`setup_configuration` method.

        :meta private:
        """
        return self.setup_configuration(**kwargs)

    def setup_configuration(
        self,
        config_dict: Optional[ConfigDictType] = None,
        check_cfg: bool = True,
        recursion: bool = True,
        substitute: bool = True,
    ) -> dict[str, Union[str, int, float, bool, None]]:
        """Defines the configuration required for running this plugin.

        Basically the default values from :class:`__parameters__` will be
        updated with the values from ``config_dict``.
        There are some special values pointing to user-related managed by the
        system defined in :class:`evaluation_system.model.user.User.getUserVarDict` .

        Parameters
        ----------
        config_dict:
            dictionary with the configuration to be used when generating the
            configuration file.
        check_cfg:
            whether the method checks that the resulting configuration
            dictionary (i.e. the default updated by `config_dict`) has no None
            values after all substituions are made.
        recursion:
            Whether when resolving the template recursion will be applied,
            i.e. variables can be set with the values of other variables,
            e.g. ``recursion && a==1 && b=="x${a}x" => f(b)=="x1x"``

        Returns
        -------
        dict:
            a copy of self.self.__config_metadict__ with all defaults values
            plus those provided here.


        :meta private:
        """
        config_dict = config_dict or {}
        if config_dict:
            conf = dict(self.__parameters__)
            conf.update(config_dict)
            config_dict = conf
        else:
            config_dict = dict(self.__parameters__)
        if substitute:
            results = self._special_variables.substitute(
                config_dict, recursive=recursion
            )
        else:
            results = config_dict.copy()

        if check_cfg:
            self.__parameters__.validate_errors(results, raise_exception=True)

        return results

    def read_from_config_parser(self, config_parser: ConfigParser) -> dict[str, str]:
        """Reads a configuration from a config parser object.

        The values are assumed to be in a section named just like the
        class implementing this method.

        Parameters
        ----------
        config_parser:
            From where the configuration is going to be read.

        Returns
        -------
        dict[str, str]:
            Updated copy of :class:`__config_metadict__`

        :meta private:
        """

        section = self.__class__.__name__
        # create a copy of metadict
        result = dict(self.__parameters__)
        # we do this to avoid having problems with the "DEFAULT" section as it
        # might define
        # more options that what this plugin requires
        keys = set(result).intersection(config_parser.options(section))
        # update values as found in the configuration
        for key in keys:
            # parse the value as good as possible
            result[key] = self.parse_config_str_value(
                key, config_parser.get(section, key)
            )
        return result

    @deprecated_method("PluginAbstract", "read_configuration")
    def readConfiguration(self, **kwargs) -> dict[str, str]:  # pragma: no cover
        """Deprecated version of the :class:`read_configuration` method.

        :meta private:
        """
        return self.read_configuration(**kwargs)

    def read_configuration(self, fp: Iterable[str]) -> dict[str, str]:
        """Read the configuration from a file object using a ConfigParser.

        Parameters
        ----------
        fp:
            File descriptor pointing to the file where the configuration is stored.

        Returns
        -------
        dict : Updated copy of :class:`__config_metadict__`

        :meta private:
        """
        config_parser = ConfigParser(interpolation=ExtendedInterpolation())
        config_parser.read_file(fp)
        return self.read_from_config_parser(config_parser)

    def save_configuration(
        self,
        fp: Union[TextIO, IO[str]],
        config_dict: Optional[ConfigDictType] = None,
        include_defaults: bool = False,
    ) -> IO[str]:
        """Stores the given configuration to the provided file object.

        If no configuration is provided the default one will be used.

        Parameters
        ----------
        fp:
            File descriptor pointing to the file where the configuration is stored.
        config_dict:
            a metadict with the configuration to be stored. If none is provided
            the result from :class:`setup_configuration`
            with ``check_cfg=False`` will be used.
        include_defaults:
            include the default parameters.

        Returns
        -------
        typing.IO: Open file descriptor, pointing to the file where the config
                   is stored

        :meta private:
        """
        # store the section header
        if config_dict is None:
            # a default incomplete one
            config_dict = self.setup_configuration(check_cfg=False, substitute=False)
        fp.write("[%s]\n" % self.__class__.__name__)
        wrapper = textwrap.TextWrapper(
            width=80,
            initial_indent="#: ",
            subsequent_indent="#:  ",
            replace_whitespace=False,
            break_on_hyphens=False,
            expand_tabs=False,
        )
        # preserve order
        for param_name in self.__parameters__:
            if include_defaults:
                param = self.__parameters__.get_parameter(param_name)
                if param.help:
                    # make sure all new lines are comments!
                    help_lines = param.help.splitlines()
                    if param.mandatory:
                        help_lines[0] = "[mandatory] " + help_lines[0]
                    fp.write("\n".join([wrapper.fill(line) for line in help_lines]))
                    fp.write("\n")
                value = config_dict.get(param_name, None)
                if value is None:
                    # means this is not setup
                    if param.mandatory:
                        value = "<THIS MUST BE DEFINED!>"
                    else:
                        value = param.default
                        param_name = "#" + param_name
                fp.write("%s=%s\n\n" % (param_name, value))
                fp.flush()  # in case we want to stream this
            elif param_name in config_dict:
                param = self.__parameters__.get_parameter(param_name)
                value = config_dict[param_name]
                key_help = param.help
                if key_help:
                    # make sure all new lines are comments!
                    help_lines = key_help.splitlines()
                    if param.mandatory:
                        help_lines[0] = "[mandatory] " + help_lines[0]
                    fp.write("\n".join([wrapper.fill(line) for line in help_lines]))
                    fp.write("\n")
                if value is None:
                    # means this is not setup
                    if param.mandatory:
                        value = "<THIS MUST BE DEFINED!>"
                    else:
                        value = ""
                        param_name = "#" + param_name
                fp.write("%s=%s\n\n" % (param_name, param.to_str(value)))
                fp.flush()  # in case we want to stream this
        return fp

    def suggest_batchscript_name(self) -> str:
        """
        Return a suggestion for the batch script file name

        Returns
        -------
        str:
            file name of the batch mode script

        :meta private:
        """

        filename = (
            datetime.now().strftime("%Y%m%d_%H%M%S_")
            + self.__class__.__name__
            + "_"
            + str(self.rowid)
        )
        return filename

    def submit_job_script(
        self,
        config_dict: Optional[ConfigDictType] = None,
        user: Optional[User] = None,
        scheduled_id: Optional[int] = None,
        log_directory: Optional[str] = None,
        unique_output: bool = True,
        extra_options: Optional[list[str]] = None,
    ) -> tuple[int, str]:
        """Create a job script suitable for the configured workload manager.

        Parameters
        ----------

        config_dict:
            Dictionary holding the plugin config setup
        user:
            Django user object
        scheduled_id:
            The row-id of a scheduled job in history
        unique_output:
            flag indicating if the output directory should be unique
        extra_options:
            Extra options passed to the workload manager, batchmode only

        Returns
        -------
        tuple[int, str]:
            The workload manager job id, the file containing the std out.

        :meta private:
        """
        log_directory = log_directory or tempfile.mkdtemp()
        if user is None:
            user = self.user
        if scheduled_id:
            cmd = self.compose_command(
                scheduled_id=scheduled_id, unique_output=unique_output
            )
        else:
            cmd = self.compose_command(
                config_dict=config_dict or {}, unique_output=unique_output
            )

        cfg = config.get_section("scheduler_options").copy()
        cfg["args"] = cmd
        cfg["extra_options"] = extra_options or []
        cfg["name"] = self.__class__.__name__
        return schedule_job(
            config.get("scheduler_system"),
            Path(config.CONFIG_FILE).parent / "activate_sh",
            cfg,
            delete_job_script=log.root.level <= logging.DEBUG,
            log_directory=log_directory,
            config_file=Path(config.CONFIG_FILE),
        )

    class ExceptionMissingParam(Exception):
        """
        An exception class if a mandatory parameter has not been set

        :meta private:
        """

        def __init__(self, param: str):
            """
            Exceptions constructor

            Parameters
            -----------
            param:
                The missing parameter
            """
            Exception.__init__(self, "Parameter %s has to be set" % param)

    def compose_command(
        self,
        config_dict: Optional[ConfigDictType] = None,
        scheduled_id: Optional[int] = None,
        batchmode: bool = False,
        email: Optional[str] = None,
        caption: Optional[str] = None,
        unique_output: bool = True,
    ) -> list[str]:
        """Create the plugin command that is submitted.

        :meta private:
        """
        logging.debug("config dict:" + str(config_dict))
        logging.debug("scheduled_id:" + str(scheduled_id))
        tool_param = [self.__class__.__name__.lower()]
        cmd_param = []
        # write explicitly if batchmode is requested
        if batchmode:
            cmd_param.append("--batchmode")
        # add a given e-mail
        if email:
            cmd_param.append(f"--mail={email}")
        # add a caption if given
        if caption is not None:
            quote_caption = caption
            quote_caption = caption.replace("\\", "\\\\")
            quote_caption = quote_caption.replace("'", "'\\''")
            cmd_param.append(f"--caption '{quote_caption}'")

        # append the unique_output param
        cmd_param.append(f"--unique_output {unique_output}")

        # a scheduled id overrides the dictionary behavior
        if scheduled_id:
            cmd_param.append(f"--scheduled-id {scheduled_id}")

        else:
            # store the section header
            if config_dict is None:
                # a default incomplete one
                config_dict = self.setup_configuration(
                    check_cfg=False, substitute=False
                )
            else:
                config_dict = self.setup_configuration(
                    config_dict=config_dict, check_cfg=False, substitute=False
                )

            # compose the parameters preserve order
            for param_name in self.__parameters__:
                if param_name in config_dict:
                    param = self.__parameters__.get_parameter(param_name)
                    value = config_dict[param_name]
                    isMandatory = param.mandatory

                if value is None and param_name not in config.exclude:
                    if isMandatory:
                        raise self.ExceptionMissingParam(param_name)
                elif param_name not in config.exclude:
                    tool_param.append(f"{param_name}={param.to_str(value)}")
        logging.debug(f"Execute command: {' '.join(tool_param+cmd_param)}")
        return tool_param + cmd_param

    def call(
        self,
        cmd_string: Union[str, list[str]],
        check: bool = True,
        **kwargs,
    ) -> sub.Popen[Any]:
        """Run command with arguments and return a CompletedProcess instance.

        The returned instance will have attributes args, returncode, stdout and
        stderr. By default, stdout and stderr are not captured, and those
        attributes will be None. Pass stdout=PIPE and/or stderr=PIPE in order
        to capture them.

        Please refer to the ``subprocess.Popen`` python build in module for
        more details.

        Parameters
        ----------
        cmd_string:
            the command to be submitted.
        check:
            If check is True and the exit code was non-zero, it raises a
            CalledProcessError. The CalledProcessError object will have the
            return code in the returncode attribute, and output & stderr
            attributes if those streams were captured.
        kwargs:
            Additional arguments passed to ``subprocess.Popen``

        Returns
        -------
        subprocess.Popen:
            sub-process return value
        """
        if isinstance(cmd_string, str):
            cmd = shlex.split(cmd_string)
        else:
            cmd = cmd_string
        log.debug("Calling: %s", " ".join(cmd))
        try:
            res = self._execute(cmd, check=check, **kwargs)
        except sub.CalledProcessError as cmd_error:
            with hide_exception():
                raise cmd_error
        return res

    def _split_path(self, path: str) -> list[str]:
        """Help function to split a path"""
        rest_path = os.path.normpath(path)
        result: list[str] = []
        while rest_path:
            old_path = rest_path
            rest_path, path_item = os.path.split(rest_path)
            if old_path == rest_path:
                result.insert(0, rest_path)
                break
            result.insert(0, path_item)
        return result



"""Module to print command argument line argument choices."""

import sys
import logging
from .utils import print_choices, logger

logger.setLevel(logging.ERROR)

try:
    print_choices(sys.argv[1:])
except KeyboardInterrupt:
    pass



#!/bin/env python
"""
This files encapsulates access to the esgf p2p system.
"""
import json
import urllib.request
import sys
import logging
import copy
from evaluation_system.misc.utils import Struct

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class P2P(object):
    """
    Handle the connection to a P2P over the Search API
    (check the `Search API documentation <http://esgf.org/wiki/ESGF_Search_API>`_.
    """

    TYPE = Struct.from_dict({"DATASET": "Dataset", "FILE": "File"})
    """Types of results that will be returned."""

    __MAX_RESULTS = 1000
    """Maximum number of items returned per call."""

    _time_out = 30
    """Specifies the time out (seconds) when trying to reach an index node."""

    def __init__(
        self,
        node="esgf-data.dkrz.de",
        api="esg-search/search",
        wget_api="esg-search/wget",
        defaults=None,
    ):
        """Creates a connection to the P2P search API.

        :param node: the p2p search node to connect to.
        :type node: str
        :param api: the url path of the service.
        :type api: str
        :param defaults: start the connection with the given defaults used for every search.
        :type defaults: dict
        """
        self.node = node
        self.api = api
        self.wget_api = wget_api
        self.defaults = defaults or {}

    def __str__(self):  # pragma: no cover
        return "P2P Search API connected to %s" % self.node

    def __repr__(self):  # pragma: no cover
        return "Search API: %s - defaults:%s" % (self.__get_url(), self.defaults)

    def set_defaults(self, defaults):
        """Set the defaults that will be used every time the node is contacted. For example::

            set_defaults(dict(project ='CMIP5'))

        :param defaults: defaults to use.
        :type defaults: dict"""
        self.defaults = defaults

    def get_defaults(self):
        """Return the defaults being currently used."""
        return copy.deepcopy(self.defaults)

    def reset_defaults(self):
        """Clear all defaults. It's the same as::

        set_defaults({})

        ."""
        self.defaults = {}

    def add_defaults(self, **defaults):
        """Adds values to the defaults for this connection.For example::

            add_defaults(project='CMIP5', product=['output1','output2'], institute_not_='MPI-M')

        This replaces but not removes existing keys."""
        self.defaults.update(defaults)

    def del_defaults(self, *def_keys):
        """Remove all the give defaults values if found. For example::

            del_defaults('institute','model')

        This would remove all values for both *institute* and *model* leaving everything else intact."""
        for k in def_keys:
            if k in self.defaults:
                del self.defaults[k]
            elif k + "_not_" in self.defaults:
                del self.defaults[k + "_not_"]

    def duplicate(self):
        """Create a duplicate of this p2p connection.

        :returns: A new instance of :class:`P2P` with the same configuration as this one."""
        return P2P(node=self.node, api=self.api, defaults=self.get_defaults())

    def __get_url(self):
        """Just return the url to this service from the values in this object"""
        return "http://%s/%s?format=application%%2Fsolr%%2Bjson" % (self.node, self.api)

    def __constraints_to_str(self, constraints, **defaults):
        """Transform the constraints passed into a url string. Apply defaults on top of it."""
        proc_const = []
        # work on a copy
        constraint_dict = self.defaults.copy()

        # apply the object wide defaults
        if constraints:
            constraint_dict.update(constraints)

        # if a method defines some defaults apply them on top
        # note this works with positive changes, you can't remove anything (yet)
        if defaults:
            constraint_dict.update(defaults)

        # encode constraints into query
        if constraint_dict:
            for key, value in constraint_dict.items():
                # negations might have been coded as <property>_not_ if not passed in a dictionary
                # just transform them to suffix '!'
                if key[-5:] == "_not_":
                    key = key[:-5] + "!"
                if isinstance(value, list):
                    value = ("&%s=" % key).join(map(str, value))
                proc_const.append("%s=%s" % (key, value))

        return "&".join(proc_const)

    def get_wget(self, **constraints):
        """:returns: (str) a string containing the wget script that can be used for downloading the selected files"""
        query = self.__constraints_to_str(constraints, type=P2P.TYPE.FILE)
        request = "http://%s/%s?%s" % (self.node, self.wget_api, query)
        return urllib.request.urlopen(request, None, self._time_out).read()

    def raw_search(self, query):
        """A raw search to the Solr index returning everything we get back.

        :returns: (dict) the raw result as returned from Solr using the search API."""
        request = "%s&%s" % (self.__get_url(), query)

        log.debug(request)
        response = json.load(urllib.request.urlopen(request, None, self._time_out))

        return response

    def search(self, type=TYPE.DATASET, **constraints):
        """Issue a query to the  P2P Search API from the constraint passed and return just the response
        (no headers).

        :param type: type of elements to return.
        :type type: :class:`P2P.TYPE`
        :param constraints: dictionary with any P2P Search API constraints.
        :returns: all documents returned from the P2P Search API."""
        query = self.__constraints_to_str(constraints, type=type)
        return self.raw_search(query)["response"]

    def get_datasets_names(self, batch_size=__MAX_RESULTS, **constraints):
        """:returns: a list of (datasets, version) tuples. (You can't define "fields" while calling this method)."""
        if "fields" in constraints:
            del constraints["fields"]
        datasets = set(
            [
                (d["master_id"], int(d["version"]))
                for d in self.datasets(
                    fields="master_id,version", batch_size=batch_size, **constraints
                )
            ]
        )
        return datasets

    def get_datasets(self, **constraints):
        """
        :returns: a list of dataset names. There's no limit imposed in the method itself, So it will
                  run out of memory if too many datasets are tried to be retrieved. Use the constraint limit=N to get
                  just the first "N" records.
        :param constraints: dictionary with any P2P Search API constraints."""

        return [d for d in self.datasets(**constraints)]

    def datasets(self, batch_size=__MAX_RESULTS, **constraints):
        """:returns: a generator iterating through the retrieve Solr docs.
        :param batch_size: defines the size of the retrieved batch. It means a query will
                           be sent to Solr every ``batch_size`` entries from the result set.
        :type batch_size: int
        :param constraints: dictionary with any P2P Search API constraints."""
        # if we have a limit use it also here.
        max_items = sys.maxsize
        try:
            max_items = int(constraints["limit"][0])
        except KeyError:
            # if not set the query limit to the batch size
            constraints["limit"] = batch_size
        try:
            sofar = int(constraints.pop("offset")[0])
        except KeyError:
            sofar = 0

        result = self.search(offset=sofar, **constraints)

        total = min(max_items, result["numFound"])
        retrieved = len(result["docs"])
        for d in result["docs"]:
            yield d

        while sofar + retrieved < total:
            result = self.search(offset=sofar + retrieved, **constraints)
            total = min(result["numFound"], max_items)
            for d in result["docs"]:
                yield d
            sofar += retrieved
            retrieved = len(result["docs"])

    def files(self, type=TYPE.FILE, **constraints):
        """the same as :class:`P2P.datasets`, but sets the type to 'File'. It's just a commodity function to
        improve readability.

        :param constraints: dictionary with any P2P Search API constraints."""
        return self.datasets(type=type, **constraints)

    def get_facets(self, *facets, **constraints):
        """
        :param facets: list of facets to query for.
        :param constraints: dictionary with any P2P Search API constraints.
        :returns: ([{facet:{facet_value:count}}]) the facets count from the server."""
        try:
            constraints = constraints.copy()
            del constraints["limit"]
        except KeyError:
            pass

        # facets encoded as a list
        if len(facets) == 1 and isinstance(facets[0], list):
            facets = facets[0]
        query = "%s&limit=0&facets=%s" % (
            self.__constraints_to_str(constraints),
            ",".join(facets),
        )

        response = self.raw_search(query)["facet_counts"]["facet_fields"]
        result = {}
        for facet in response:
            result[facet] = dict(zip(response[facet][0::2], response[facet][1::2]))
        return result

    @staticmethod
    def show(dictio, return_str=False):
        """Pretty print json (or any dict)"""
        json_str = json.dumps(dictio, sort_keys=True, indent=2)
        if return_str:
            return json_str
        print(json_str)

    @staticmethod
    def extract_catalog(dictio):
        """Extract the catalog from this json returned info. Only 1 catalog is expected."""
        if "url" in dictio:
            result = [
                e.split(".xml")[0] + ".xml"
                for e in dictio["url"]
                if e[-7:] == "Catalog"
            ]
            if result and len(result) == 1:
                return result[0]
            else:
                raise Exception("Can't find unique catalog")



import time
import gzip
from datetime import datetime
from evaluation_system.model.esgf import P2P
from evaluation_system.misc import config
from esgf_crawl_config import Esgf2SolrConfig


class Esgf2Solr(object):
    def __init__(self, project, experiment, outpath, p2p=P2P()):
        self.project = {"project": project}
        self.experiment = experiment
        self.outpath = outpath
        self.show_facets = "experiment"

        self.facets = {"project": list(self.project.values())[0], "type": "File"}
        self.fields = [
            "title",
            "size",
            "project",
            "product",
            "institute",
            "model",
            "experiment",
            "time_frequency",
            "realm",
            "variable",
            "ensemble",
            "timestamp",
        ]
        self.prefix = "esgf-"
        self.prepath = config.get("project_data")

    def find_experiment(self, p2p=P2P()):
        results = p2p.get_facets(self.show_facets, **self.project)
        if len(results["experiment"].keys()) == 0:
            raise Exception("Experiment not found")
        self.experiments = results["experiment"].keys()

    def get_path(self, p2p=P2P()):
        gzname = self.outpath + "/solr_crawl_%s.csv.gz" % datetime.now().strftime(
            "%Y-%m-%d_%H%M%S"
        )
        try:
            with gzip.open(gzname, "wb", 6) as f:
                f.write(
                    "crawl_dir\t%s%s%s\n"
                    % (self.prepath, self.prefix, self.project["project"])
                )
                f.write("\n")
                f.write("data\n")
                for experiment in self.experiments:
                    if self.experiment is not None and experiment != self.experiment:
                        continue
                    facets = self.facets
                    facets.update({"experiment": experiment})
                    esgffiles = {}
                    for dataset in p2p.get_datasets(
                        fields=",".join(self.fields), **facets
                    ):
                        cmor_path = Esgf2SolrConfig().project_select(dataset)
                        filename = cmor_path["filename"]
                        esgfpath = "/".join(
                            (
                                self.prefix + cmor_path["project"],
                                cmor_path["product"],
                                cmor_path["institute"],
                                cmor_path["model"],
                                cmor_path["experiment"],
                                cmor_path["time_frequency"],
                                cmor_path["realm"],
                                cmor_path["cmor_var"],
                                cmor_path["ensemble"],
                            )
                        )
                        esgfpath += "/"
                        timestamp = time.mktime(
                            (
                                datetime.strptime(
                                    dataset["timestamp"], "%Y-%m-%dT%H:%M:%SZ"
                                )
                            ).timetuple()
                        )
                        if self.prepath + esgfpath + filename not in esgffiles:
                            esgffiles[self.prepath + esgfpath + filename] = 0
                        if timestamp >= esgffiles[self.prepath + esgfpath + filename]:
                            esgffiles[self.prepath + esgfpath + filename] = timestamp
                    [f.write(key + ",%s\n" % value) for key, value in esgffiles.items()]
                f.close()
        except IOError:
            raise FileNotFoundError("Path does not exist")



"""Manage the abstraction of a system user."""
from __future__ import annotations
import pwd
import os
from configparser import ConfigParser as Config, ExtendedInterpolation
from typing import Optional, Union

from evaluation_system.misc import config, utils
from evaluation_system.model.db import UserDB


class User:
    """
    This Class encapsulates a user (configurations, etc).

    Parameters
    ----------
    uid: int
        user id in the local system, if not provided the current user is used.
    email: str
        user's email address
    """

    CONFIG_DIR: str = "config"
    """The directory name where all plug-in/system configurations will be stored."""

    CACHE_DIR: str = "cache"
    """The temporary directory where plug-ins can store files while performing some computation."""

    OUTPUT_DIR: str = "output"
    """The directory where output files are stored. Intended for files containing data and thus taking much space."""

    PLOTS_DIR: str = "plots"
    """The directory where just plots are stored. Plots are assumed to be much smaller in size than data and might
therefore live longer"""

    PROCESSES_DIR: str = "processes"
    """The directory might handle information required for each running process."""

    EVAL_SYS_CONFIG: str = os.path.join(CONFIG_DIR, "evaluation_system.config")
    """The file containing a central configuration for the whole system (user-wise)"""

    EVAL_SYS_DEFAULT_CONFIG: str = os.path.normpath(
        os.path.dirname(__file__) + "/../../etc/system_default.config"
    )
    """The central default configuration file for all users. It should not be confused with the system configuration
file that is handled by :class:`evaluation_system.api.config`."""

    def __init__(self, uid: Optional[Union[int, str]] = None, email: str = ""):
        """Creates a user object for the provided id.

        If no id is given, a user object for
        the current user, i.e. the one that started the application, is created instead.
        """
        self._dir_type = config.get(config.DIRECTORY_STRUCTURE_TYPE)
        uid = uid or os.getuid()
        self._userdata = None
        if isinstance(uid, str):
            self._userdata = pwd.getpwnam(uid)
        else:
            self._userdata = pwd.getpwuid(uid)

        if self._userdata is None:
            raise Exception("Cannot find user %s" % uid)
        self._email = email
        self._userconfig = Config(interpolation=ExtendedInterpolation())
        # try to load teh configuration from the very first time.
        self._userconfig.read(
            [
                User.EVAL_SYS_DEFAULT_CONFIG,
                os.path.join(self._userdata.pw_dir, User.EVAL_SYS_CONFIG),
            ]
        )

        self._db = UserDB(self)

        row_id = self._db.getUserId(self.getName())

        if row_id:
            self._db.updateUserLogin(row_id, email)
        else:
            self._db.createUser(self.getName(), email=self._email)

    # $USER_BASE_DIR := central directory for this user in the evaluation system.
    # $USER_OUTPUT_DIR := directory where the output data for this user is stored.
    # $USER_PLOT_DIR := directory where the plots for this user is stored.
    # $USER_CACHE_DIR := directory where the cached data for this user is stored."""

    def __str__(self):
        return "<User (username:%s, info:%s)>" % (
            self._userdata[0],
            str(self._userdata[2:]),
        )

    def getUserConfig(self):
        """:returns: the user configuration object :py:class:`ConfigParser.SafeConfigParser`"""
        return self._userconfig

    def getUserDB(self):
        """:returns: the db abstraction for this user.
        :rtype: :class:`evaluation_system.model.db.UserDB`"""
        return self._db

    def reloadConfig(self):
        """Reloads user central configuration from disk (not the plug-in related one)."""
        self._userconfig = Config(interpolation=ExtendedInterpolation())
        self._userconfig.read(
            [
                User.EVAL_SYS_DEFAULT_CONFIG,
                os.path.join(self.getUserBaseDir(), User.EVAL_SYS_CONFIG),
            ]
        )
        return self._userconfig

    def writeConfig(self):
        """Writes the user central configuration to disk according to :class:`EVAL_SYS_CONFIG`"""

        fp = open(os.path.join(self.getUserBaseDir(), User.EVAL_SYS_CONFIG), "w")
        self._userconfig.write(fp)
        fp.close()

    def getName(self):
        """:returns: the user name
        :rtype: str"""
        return self._userdata.pw_name

    def getEmail(self):
        """
        :returns: user's email address. Maybe None. :rtype: str
        """
        return self._email

    def getUserID(self):
        """:returns: the user id.
        :rtype: int"""
        return self._userdata.pw_uid

    def getUserHome(self):
        """:returns: the path to the user home directory.
        :rtype: str"""
        return self._userdata.pw_dir

    def getUserScratch(self):
        """:returns: the path to the user's scratch directory.
        :rtype: str"""
        return self._getUserBaseDir()

    def _getUserBaseDir(self):
        if self._dir_type == config.DIRECTORY_STRUCTURE.LOCAL:
            return os.path.join(self.getUserHome(), config.get(config.BASE_DIR))
        else:
            return os.path.join(
                config.get(config.BASE_DIR_LOCATION),
                str(self.getName()),
                config.get(config.BASE_DIR),
            )

    def _getUserDir(self, dir_type, tool=None, create=False):
        base_dir = dict(
            base="",
            config=User.CONFIG_DIR,
            cache=User.CACHE_DIR,
            output=User.OUTPUT_DIR,
            plots=User.PLOTS_DIR,
            processes=User.PROCESSES_DIR,
            scheduler_in=config.get(config.SCHEDULER_INPUT_DIR),
            scheduler_out=config.get(config.SCHEDULER_OUTPUT_DIR),
        )

        if tool is None:
            bd = base_dir[dir_type]
            # concatenate relative paths only
            if bd and bd[0] == "/":
                dir_name = bd
            else:
                # return the directory where the tool configuration files are stored
                dir_name = os.path.join(self._getUserBaseDir(), bd)
        else:
            # It's too confusing if we create case sensitive directories...
            tool = tool.lower()
            # return the specific directory for the given tool
            dir_name = os.path.join(self._getUserBaseDir(), base_dir[dir_type], tool)

        # make sure we have a canonical path
        dir_name = os.path.abspath(dir_name)

        if create and not os.path.isdir(dir_name):
            # we are letting this fail in case of problems.
            utils.supermakedirs(dir_name, 0o0755)

        return dir_name

    def getUserBaseDir(self, **kwargs):
        """Returns path to where this system is managing this user data.

        :param kwargs: ``create`` := If ``True`` assure the directory exists after the call is done.
        :returns: (str) path"""
        return self._getUserDir("base", **kwargs)

    def getUserSchedulerInputDir(self, **kwargs):
        """Returns path to where this system is managing this user data.

        :param kwargs: ``create`` := If ``True`` assure the directory exists after the call is done.
        :returns: (str) path"""
        return self._getUserDir("scheduler_in", **kwargs)

    def getUserSchedulerOutputDir(self, **kwargs):
        """Returns path to where this system is managing this user data.

        :param kwargs: ``create`` := If ``True`` assure the directory exists after the call is done.
        :returns: (str) path"""
        return self._getUserDir("scheduler_out", **kwargs)

    def getUserToolConfig(self, tool, **kwargs):
        """Returns the path to the configuration file.

        :param kwargs: ``create`` := If ``True`` assure the underlaying directory exists after the call is done.
        :param tool: tool/plug-in for which the information is returned.
        :type tool: str
        :returns: path to the configuration file."""
        config_dir = self._getUserDir("config", tool, **kwargs)
        return os.path.join(config_dir, "%s.conf" % tool)

    def getUserConfigDir(self, tool=None, **kwargs):
        """Return the path to the directory where all configurations for this user are stored.

        :param kwargs: ``create`` := If ``True`` assure the  directory exists after the call is done.
        :param tool: tool/plug-in for which the information is returned. If None, then the directory
                     where all information for all tools reside is returned insted (normally, that would
                     be the parent directrory).
        :type tool: str
        :returns: path to the directory."""
        return self._getUserDir("config", tool, **kwargs)

    def getUserCacheDir(self, tool=None, **kwargs):
        """Return directory where cache files for this user (might not be "only" for this user though).

        :param kwargs: ``create`` := If ``True`` assure the  directory exists after the call is done.
        :param tool: tool/plug-in for which the information is returned. If None, then the directory
                     where all information for all tools reside is returned insted (normally, that would
                     be the parent directrory).
        :type tool: str
        :returns: path to the directory."""
        return self._getUserDir("cache", tool, **kwargs)

    def getUserProcessDir(self, tool=None, **kwargs):
        """Return directory where files required for processes can be held. Is not clear what this will
        be used for, but it should at least serve as a possibility for the future.

        :param kwargs: ``create`` := If ``True`` assure the  directory exists after the call is done.
        :param tool: tool/plug-in for which the information is returned. If None, then the directory
                     where all information for all tools reside is returned insted (normally, that would
                     be the parent directrory).
        :type tool: str
        :returns: path to the directory."""
        return self._getUserDir("processes", tool, **kwargs)

    def getUserOutputDir(self, tool=None, **kwargs):
        """Return directory where output data for this user is stored.

        :param kwargs: ``create`` := If ``True`` assure the  directory exists after the call is done.
        :param tool: tool/plug-in for which the information is returned. If None, then the directory
                     where all information for all tools reside is returned insted (normally, that would
                     be the parent directrory).
        :type tool: str
        :returns: path to the directory."""
        return self._getUserDir("output", tool, **kwargs)

    def getUserPlotsDir(self, tool=None, **kwargs):
        """Return directory where all plots for this user are stored.

        :param kwargs: ``create`` := If ``True`` assure the  directory exists after the call is done.
        :param tool: tool/plug-in for which the information is returned. If None, then the directory
                     where all information for all tools reside is returned insted (normally, that would
                     be the parent directrory).
        :type tool: str
        :returns: path to the directory."""
        return self._getUserDir("plots", tool, **kwargs)

    def prepareDir(self):
        """Prepares the configuration directory for this user if it's not already been done."""
        if os.path.isdir(self.getUserBaseDir()):
            # we assume preparation was successful... but we might to be sure though...
            # return
            pass

        if not os.path.isdir(self.getUserHome()):
            raise Exception(
                "Can't create configuration, user HOME doesn't exist (%s)"
                % self.getUserHome()
            )

        # create directory for the framework
        # create all required subdirectories
        dir_creators = [
            self.getUserBaseDir,
            self.getUserConfigDir,
            self.getUserCacheDir,
            self.getUserOutputDir,
            self.getUserPlotsDir,
            self.getUserSchedulerInputDir,
            self.getUserSchedulerOutputDir,
        ]

        for f in dir_creators:
            f(create=True)



"""
Created on 31.05.2016

@author: Sebastian Illing
"""
import os
from pathlib import Path
import tempfile
import shutil
import logging
import time
import textwrap
import datetime
from tempfile import TemporaryDirectory
import pytest

from evaluation_system.api.plugin_manager import PluginMetadata

log = logging.getLogger(__name__)


def test_modules(dummy_settings):
    import evaluation_system.api.plugin_manager as pm

    pm.reload_plugins()
    pmod = pm.__plugin_modules_user__
    assert pmod is not None
    assert len(pmod) > 0


def test_plugins(dummy_settings, temp_user):
    from evaluation_system.tests.mocks.dummy import DummyPlugin
    import evaluation_system.api.plugin_manager as pm

    # force reload to be sure the dummy is loaded
    assert len(pm.get_plugins()) > 0
    assert "dummyplugin" in pm.get_plugins()
    dummy = pm.get_plugin_metadata("dummyplugin")
    assert dummy.description == DummyPlugin.__short_description__
    assert dummy.category == DummyPlugin.__category__
    assert dummy.tags == DummyPlugin.__tags__
    assert dummy.plugin_class == "DummyPlugin"
    os.environ[
        f"EVALUATION_SYSTEM_PLUGINS_{temp_user.getName()}"
    ] = f"{str(Path(__file__).parent / 'mocks')},dummy"
    pm.reload_plugins(temp_user.getName())
    os.environ.pop(f"EVALUATION_SYSTEM_PLUGINS_{temp_user.getName()}")


def testDefaultPluginConfigStorage(temp_user):
    import evaluation_system.api.plugin_manager as pm

    pm.reload_plugins(temp_user.username)
    home = temp_user.getUserHome()
    assert os.path.isdir(home)
    conf_file = pm.write_setup("dummyplugin", user=temp_user)
    assert os.path.isfile(conf_file)


def test_plugin_config_storage(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm
    from evaluation_system.api.parameters import ValidationError

    home = temp_user.getUserHome()
    assert os.path.isdir(home)

    res = pm.get_plugin_instance("dummyplugin").setup_configuration(
        config_dict=dict(the_number=42)
    )
    assert res["something"] == "test"

    # write down this default
    conf_file = pm.write_setup(
        "dummyplugin", config_dict=dict(the_number=42), user=temp_user
    )

    assert os.path.isfile(conf_file)
    with open(conf_file, "r") as f:
        config = f.read()

    assert "\nsomething=test\n" in config

    with pytest.raises(ValidationError):
        pm.parse_arguments("dummyplugin", [])
    with pytest.raises(ValidationError):
        pm.parse_arguments("dummyplugin", [], user=temp_user)
    res = pm.parse_arguments("dummyplugin", [], use_user_defaults=True, user=temp_user)

    assert res == {
        "other": 1.4,
        "number": None,
        "the_number": 42,
        "something": "test",
        "input": None,
        "variable": "tas",
        "extra_scheduler_options": "",
    }
    assert res["something"] == "test"

    # now change the stored configuration
    config = config.replace("\nsomething=test\n", "\nsomething=super_test\n")
    with open(conf_file, "w") as f:
        f.write(config)
    res = pm.parse_arguments("dummyplugin", [], use_user_defaults=True, user=temp_user)
    assert res["something"] == "super_test"


def test_parse_arguments(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm

    home = temp_user.getUserHome()
    assert os.path.isdir(home)

    # direct parsing
    for args, result in [
        (
            "the_number=4",
            {
                "other": 1.3999999999999999,
                "the_number": 4,
                "something": "test",
                "variable": "tas",
                "extra_scheduler_options": "",
            },
        )
    ]:
        d = pm.parse_arguments("Dummyplugin", args.split(), user=temp_user)
        assert d == result

    # parsing requesting user default but without any
    for args, result in [
        (
            "the_number=4",
            {
                "other": 1.3999999999999999,
                "the_number": 4,
                "something": "test",
                "variable": "tas",
                "extra_scheduler_options": "",
            },
        )
    ]:
        d = pm.parse_arguments("Dummyplugin", args.split(), user=temp_user)
        assert d == result

    pm.write_setup("DummyPlugin", dict(number=7, the_number=42), temp_user)
    for args, result in [
        (
            "number=4",
            dict(
                number=4,
                the_number=42,
                something="test",
                other=1.4,
                input=None,
                variable="tas",
                extra_scheduler_options="",
            ),
        )
    ]:
        d = pm.parse_arguments(
            "Dummyplugin", args.split(), use_user_defaults=True, user=temp_user
        )
        assert d == result

    shutil.rmtree(home)


def test_write_setup(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm

    f = pm.write_setup(
        "DummyPlugin", dict(number="$the_number", the_number=42), temp_user
    )

    with open(f) as fp:
        num_line = [
            line for line in fp.read().splitlines() if line.startswith("number")
        ][0]
        assert num_line == "number=$the_number"


def test_get_history(dummy_settings, temp_user):
    import evaluation_system.api.plugin_manager as pm

    pm.write_setup("DummyPlugin", dict(the_number=777), temp_user)
    pm.run_tool("dummyplugin", user=temp_user)
    res = pm.get_history(user=temp_user)
    res = res[0]
    import re

    mo = re.search(
        "^([0-9]{1,})[)] ([^ ]{1,}) ([^ ]{1,}) ([^ ]{1,})", res.__str__(compact=False)
    )
    assert mo is not None
    g1 = mo.groups()
    assert all([g is not None for g in g1])
    mo = re.search("^([0-9]{1,})[)] ([^ ]{1,}) ([^ ]{1,})", res.__str__())
    g2 = mo.groups()
    assert all([g is not None for g in g2])
    assert g1[0] == g2[0]


def testDynamicPluginLoading(dummy_env, temp_user):
    import evaluation_system.api.plugin_manager as pm

    basic_plugin = textwrap.dedent(
        """
        from sys import modules
        plugin = modules['evaluation_system.api.plugin']
        parameters = modules['evaluation_system.api.parameters']

        class %s(plugin.PluginAbstract):
            __short_description__ = "Test"
            __version__ = (0,0,1)
            __category__ = "foo"
            __tags__ = ["bar"]
            __parameters__ =  parameters.ParameterDictionary(
                                            parameters.File(name="output", default="/tmp/file", help='output'),
                                            parameters.File(name="input", mandatory=True, help="some input"))

            def runTool(self, config_dict=None):
                print("%s", config_dict)
        """
    )

    with tempfile.TemporaryDirectory(prefix="dyn_plugin") as path1:
        os.makedirs(os.path.join(path1, "a/b"))
        with open(path1 + "/a/__init__.py", "w"):
            pass
        with open(path1 + "/a/blah.py", "w") as f:
            f.write(basic_plugin % tuple(["TestPlugin1"] * 2))

        with tempfile.TemporaryDirectory(prefix="dyn_plugin") as path2:

            os.makedirs(os.path.join(path2, "x/y/z"))
            with open(path2 + "/x/__init__.py", "w"):
                pass
            with open(path2 + "/x/foo.py", "w") as f:
                f.write(basic_plugin % tuple(["TestPlugin2"] * 2))

            os.environ[pm.PLUGIN_ENV] = "%s,%s:%s,%s" % (
                "~/../../../../../.." + path1 + "/a",
                "blah",  # test a relative path starting from ~
                "$HOME/../../../../../.." + path2 + "/x",
                "foo",
            )  # test a relative path starting from $HOME
            log.debug("pre-loading: %s", list(pm.get_plugins()))

            assert "testplugin1" not in list(pm.get_plugins())
            assert "testplugin2" not in list(pm.get_plugins())
            pm.reload_plugins()
            log.debug("post-loading: %s", list(pm.get_plugins()))
            assert "testplugin1" in list(pm.get_plugins())
            assert "testplugin2" in list(pm.get_plugins())


def test_load_invalid_plugin(dummy_env, temp_user):
    import evaluation_system.api.plugin_manager as pm

    # this intentionally does not define the required properties and methods
    basic_plugin = textwrap.dedent(
        """
        from sys import modules
        plugin = modules['evaluation_system.api.plugin']
        parameters = modules['evaluation_system.api.parameters']

        class TestPlugin(plugin.PluginAbstract):
            pass
        """
    )

    with tempfile.TemporaryDirectory(prefix="dyn_plugin") as plugin_dir:
        os.makedirs(os.path.join(plugin_dir, "a"))
        with open(plugin_dir + "/a/__init__.py", "w"):
            pass
        with open(plugin_dir + "/a/test_plugin.py", "w") as f:
            f.write(basic_plugin)

        os.environ[pm.PLUGIN_ENV] = "%s,%s" % (
            plugin_dir + "/a",
            "test_plugin",  # test a relative path starting from ~
        )
        log.debug("pre-loading: %s", list(pm.get_plugins()))
        assert "testplugin" not in list(pm.get_plugins())
        pm.reload_plugins()
        log.debug("post-loading: %s", list(pm.get_plugins()))
        assert "testplugin" not in list(pm.get_plugins())


def test_get_plugin_dict(dummy_env):
    import evaluation_system.api.plugin_manager as pm

    with pytest.raises(pm.PluginManagerException):
        pm.get_plugin_metadata("Not available")
    pl = pm.get_plugin_metadata("DummyPlugin")
    assert isinstance(pl, PluginMetadata)
    assert pl.plugin_class == "DummyPlugin"


def test_preview_generation(dummy_env):
    import evaluation_system.api.plugin_manager as pm
    import evaluation_system.misc.config as config

    with tempfile.TemporaryDirectory() as td:
        d = str(Path(td) / "tmp.pdf")
        s = os.path.dirname(__file__) + "/test_output/vecap_test_output.pdf"
        pm._preview_copy(s, d)
        assert os.path.isfile(d)
    with tempfile.TemporaryDirectory() as td:
        d = str(Path(td) / "tmp.png")
        s = os.path.dirname(__file__) + "/test_output/test_image.png"
        f = open(d, "w")
        f.close()
        pm._preview_copy(s, d)
        assert os.path.isfile(d)
    with tempfile.TemporaryDirectory() as td:
        d = str(Path(td) / "tmp.png")
        s = os.path.dirname(__file__) + "/test_output/test_image.png"
        pm._preview_convert(s, d)
        assert os.path.isfile(d)

    r = pm._preview_generate_name("dummy", {})
    assert "dummy" in r
    assert len(r) == 14
    ts = time.time()
    r = pm._preview_generate_name("murcss", {"timestamp": ts})
    assert "murcss" in r
    assert datetime.datetime.fromtimestamp(ts).strftime("%Y%m%d_%H%M%S") in r

    u = pm._preview_unique_file("murcss", "pdf", {"timestamp": ts})
    assert datetime.datetime.fromtimestamp(ts).strftime("%Y%m%d_%H%M%S") in u
    assert "murcss" in u
    assert config.get("preview_path") in u

    r1 = os.path.dirname(__file__) + "/test_output/vecap_test_output.pdf"
    r2 = os.path.dirname(__file__) + "/test_output/test_image.png"
    result = {r1: {"todo": "copy"}, r2: {"todo": "convert"}}
    res = pm._preview_create("murcss", result)
    for r in res:
        assert os.path.isfile(r)
        os.remove(r)


def test_get_command_string(dummy_env, django_user):
    from evaluation_system.model.history.models import History
    import evaluation_system.api.plugin_manager as pm

    h = History.objects.create(
        timestamp=datetime.datetime.now(),
        status=History.processStatus.running,
        uid=django_user,
        configuration='{"some": "config", "dict": "values"}',
        tool="dummytool",
        slurm_output="/path/to/slurm-44742.out",
    )

    cmd = pm.get_command_string(h.id)
    assert "freva-plugin" in cmd
    assert h.tool in cmd


def test_load_scheduled_conf(dummy_env, django_user, temp_user):
    from evaluation_system.model.history.models import History
    import evaluation_system.api.plugin_manager as pm

    h = History.objects.create(
        timestamp=datetime.datetime.now(),
        status=History.processStatus.scheduled,
        uid=django_user,
        configuration='{"some": "config", "dict": "values"}',
        tool="dummytool",
        slurm_output="/path/to/slurm-44742.out",
    )

    res = pm.load_scheduled_conf("dummytool", h.id, temp_user)
    assert res == {}


def test_2dict_to_conf(dummy_env, dummy_plugin):

    import evaluation_system.api.plugin_manager as pm

    configuration = {"number": 1, "the_number": 2, "something": "test"}
    with pytest.raises(pm.PluginManagerException):
        pm.dict2conf("dummytool", configuration)
    with pytest.raises(pm.ParameterNotFoundError):
        pm.dict2conf("dummyplugin", {"some": "config"})
    dd = pm.dict2conf("dummyplugin", configuration)
    assert isinstance(dd, list)
    assert len(dd) == len(configuration)


def test_scheduletool(dummy_env, dummy_plugin):
    import evaluation_system.api.plugin_manager as pm

    with TemporaryDirectory() as td:
        pm.schedule_tool("dummyplugin", log_directory=str(Path(td) / "tmp"))



import logging
import os
import sys

logging.basicConfig(format="%(name)s - %(levelname)s - %(message)s")
logger: logging.Logger = logging.getLogger("freva")
logger.setLevel(logging.INFO)

_DEFAULT_CONFIG_FILE_LOCATION = os.path.join(
    sys.prefix, "freva", "evaluation_system.conf"
)
# now check if we have a configuration file, and read the defaults from there
CONFIG_FILE = os.environ.get(
    "EVALUATION_SYSTEM_CONFIG_FILE", _DEFAULT_CONFIG_FILE_LOCATION
)



"""
freva -- Free University Evaluation System

@copyright:  2016 FU Berlin. All rights reserved.

@contact:    sebastian.illing@met.fu-berlin.de

@license:    BSD

Copyright (c) 2016, FU Berlin
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.
    Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import importlib
from evaluation_system import __version__

__all__ = [
    "crawl_my_data",
    "databrowser",
    "run_plugin",
    "list_plugins",
    "plugin_doc",
    "read_plugin_cache" "esgf",
    "history",
]


def __getattr__(name):
    if name in ["crawl_my_data", "databrowser", "esgf", "history"]:
        return getattr(importlib.import_module(f"._{name}", __name__), name)
    elif name in [
        "run_plugin",
        "list_plugins",
        "plugin_doc",
        "get_tools_list",
        "read_plugin_cache",
    ]:
        return getattr(importlib.import_module(f"._plugin", __name__), name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""Module to encapsulates the access to repositories."""
from __future__ import annotations
from contextlib import contextmanager
from pathlib import Path
import os
from typing import Iterator

import git
from evaluation_system.misc import logger

__version_cache: dict[str, tuple[str, str]] = {}


@contextmanager
def set_repo_safe(dir_name: os.PathLike) -> Iterator[git.Repo]:
    """Context manager that sets a given repository dir as temporarly safe.

    This needs to be done to not make git complain about the fact that the
    repositories directory doesn't belong to the person who is trying to
    retrieve the repository information. Before we instruct git to read the
    repo we tell it that his directory is save by temporarly adding a safe
    flag to the global git settings. After the repo has been processed by git
    this safe flag is removed again. This is probaply not the best way of
    dealing with this issue. Still, for example setting repositories to as
    shared did not solve the issue.

    Parameters
    ----------
    dir_name: os.PathLike
        Path to the repository dir that is set to be safe

    Returns
    -------
    Iterator[git.Repo]:
        Instance of the gitpython Repo class
    """
    repo = git.Repo(dir_name, search_parent_directories=True)

    with repo.config_reader("global") as config_r:
        safe_dirs = {}
        if config_r.has_section("safe"):
            safe_dirs = dict(config_r.items_all("safe"))
    with repo.config_writer("global") as config_w:
        try:
            config_w.add_value("safe", "directory", str(dir_name)).release()
            yield repo
        except Exception as error:
            raise error
        finally:
            config_w.remove_section("safe")
            for key, values in safe_dirs.items():
                for value in values:
                    config_w.add_value("safe", key, value)


def get_version(src_file: str) -> tuple[str, str]:
    """Acquire git remote repo url and last commit hash.

    Parameters
    ----------
    src_file: str
        Path to the plugin wrapper API file

    Returns
    -------
    tuple[str, str]
        tuple holding remote repo url, last git commit hash

    """
    retval = __version_cache.get(src_file, None)
    if retval is None:
        try:
            with set_repo_safe(Path(src_file).parent) as repo:
                repository = next(repo.remote(name="origin").urls)
                version = repo.head.object.hexsha
        except Exception as error:
            logger.error("Could not read git version")
            logger.error("Error while reading git version:\n%s", str(error))
            repository = "unknown"
            version = "unknown"
        retval = (repository, version)
        __version_cache[str(src_file)] = retval
    return retval



import os
import pytest
import mock

SUBCOMMANDS = ("databrowser", "esgf", "crawl-my-data", "history", "plugin")


def test_main_complete(dummy_env, capsys):

    from freva.cli.utils import print_choices

    print_choices(["freva"])
    sub_commands = capsys.readouterr().out
    for subcommand in SUBCOMMANDS:
        assert subcommand in sub_commands
    print_choices(["--shell", "tcsh", "freva"])
    assert capsys.readouterr().out == sub_commands
    print_choices(["--shell", "zsh", "freva"])
    sub_commands_zsh = capsys.readouterr().out.split("\n")
    for complete in [s.strip() for s in sub_commands_zsh if s.strip()]:
        cmd = complete.split("[")
        assert cmd[0] in SUBCOMMANDS
        assert len(cmd) == 2
    print_choices(["--shell", "fish", "freva"])
    sub_commands_fish = capsys.readouterr().out.split("\n")
    for complete in [s.strip() for s in sub_commands_fish if s.strip()]:
        cmd = complete.split(":")
        assert cmd[0] in SUBCOMMANDS
        assert len(cmd) == 2


def test_subcommand_help(dummy_env, capsys):

    from freva.cli.utils import print_choices
    from freva.cli import main as main_cli

    for subcmd in SUBCOMMANDS:
        with pytest.raises(SystemExit):
            main_cli([subcmd, "--help"])
        doc_string = capsys.readouterr().out
        for shell, sep in {"bash": "", "zsh": "[", "fish": ":"}.items():
            print_choices(["--shell", shell, "--flags-only", f"freva-{subcmd}"])
            choices = capsys.readouterr().out.split("\n")
            for choice in [s.strip() for s in choices if s.strip()]:
                if sep:
                    cmd, help = choice.split(sep)
                    help = help.rstrip("]")
                else:
                    cmd = choice
                    help = ""
                assert cmd in doc_string
                assert cmd.startswith("-")
                if help:
                    # The 20 first characters should be ok for this comparison
                    assert help[:20] in doc_string
            if subcmd not in ("databrowser", "plugin"):
                print_choices(["--shell", shell, "--strip", f"freva-{subcmd}"])
                out = [
                    s.strip() for s in capsys.readouterr().out.split("\n") if s.strip()
                ]
                if subcmd == "crawl-my-data":
                    assert len(out) == 1
                else:
                    assert len(out) == 0


def test_databrowser(dummy_env, capsys, dummy_solr):

    from freva.cli.utils import print_choices
    from freva.cli import main as main_cli

    main_cli(["databrowser", "--all-facets"])
    facets = capsys.readouterr().out
    for shell in ("bash", "zsh", "fish"):
        print_choices(["--shell", shell, "--strip", "freva", "databrowser"])
        choices = capsys.readouterr().out
        assert choices == facets
        print_choices(
            ["--shell", shell, "--strip", "freva", "databrowser", "time_frequency=mon"]
        )
        choices = capsys.readouterr().out
        assert "time_frequency" not in choices


def test_plugin(dummy_env, capsys):

    from freva.cli.utils import print_choices
    from freva._plugin import _write_plugin_cache
    from freva.cli import main as main_cli

    _write_plugin_cache()

    for shell in ("bash", "zsh", "fish"):
        print_choices(["--shell", shell, "--strip", "freva", "plugin"])
        choices = capsys.readouterr().out
        assert "dummyplugin" in choices
        print_choices(["--shell", shell, "--strip", "freva", "plugin", "dummyplugin"])
        choices = capsys.readouterr().out
        assert "the_number" in choices
        print_choices(
            [
                "--shell",
                shell,
                "--strip",
                "freva",
                "plugin",
                "dummyplugin",
                "the_number=1",
            ]
        )
        choices = capsys.readouterr().out
        assert "the_number" not in choices


def test_wrong_choice(dummy_env, capsys):

    from freva.cli.utils import print_choices
    from freva.cli import main as main_cli

    for shell in ("bash", "zsh", "fish"):
        for cmd in (["solr"], ["plugin", "nplg"]):
            print_choices(["--shell", shell, "--strip", "freva"] + cmd)
            choices = capsys.readouterr().out
            assert len(choices) == 0



"""This module queries the database for plugin history entries."""
from __future__ import annotations
import json
from typing import Optional, Union

import evaluation_system.api.plugin_manager as pm
from evaluation_system.misc import logger

__all__ = ["history"]


def history(
    *args: str,
    limit: int = 10,
    plugin: Optional[str] = None,
    since: Optional[str] = None,
    until: Optional[str] = None,
    entry_ids: Union[int, list[int]] = None,
    full_text: bool = False,
    return_command: bool = False,
    _return_dict: bool = True,
) -> Union[list[str], dict[str, str]]:
    """Get access to the configurations of previously applied freva plugins.

    The `.history` method displays the entries with a one-line compact description.
    The first number you see is the entry id, which you might use to select
    single entries.

    Parameters
    ----------
    full_text: bool, default: False
      Get the full configuration.
    return_command: bool, default: False
      Show Freva commands belonging to the history entries instead
      of the entries themself.
    limit: int, default: 10
      Limit the number of entires to be displayed.
    plugin: str, default: None
      Display only entries from a given plugin name.
    since: str, datetime.datetime, default: None
      Retrieve entries older than date, see hint on date format below.
    until: str, datetime.datetime, default: None
      Retrieve entries younger than date, see hint on date format below.
    entry_ids: list, default: None
       Select entries whose ids are in "ids",
    full_text: bool, default: False
      Show the complete configuration.
    return_command: bool, default: False
      Return the commands instead of history objects
    _return_dict: bool, default: True
      Return a dictionary representation, this is only for internal use

    Returns
    -------
      list:
        freva plugin history

    Example
    -------

    Get the last three history entries

    .. execute_code::

        import freva
        hist = freva.history(limit=3)
        print(type(hist), len(hist))
        print(hist[-1].keys())
        config = hist[-1]['configuration']
        print(config)


    .. hint:: Date Format
        Dates can be given in "YYYY-MM-DD HH:mm:ss.n" or any less accurate subset.
        These are all valid: "2012-02-01 10:08:32.1233431", "2012-02-01 10:08:32",
        "2012-02-01 10:08", "2012-02-01 10", "2012-02-01", "2012-02", "2012".
        Missing values are assumed to be the minimal allowed value. For example:
        "2012" = "2012-01-01 00:00:00.0"
    """
    if not isinstance(entry_ids, (list, tuple, set)) and entry_ids is not None:
        entry_ids = [entry_ids]
    if not return_command:
        prefix = "history"
        if plugin:
            prefix = "history of {plugin}"
        logger.info(
            f"{prefix}, limit={limit}, since={since},"
            f" until={until}, entry_ids={entry_ids}"
        )
    rows = pm.get_history(
        user=None,
        plugin_name=plugin,
        limit=limit,
        since=since,
        until=until,
        entry_ids=entry_ids,
    )
    if rows:
        # pass some option for generating the command string
        if return_command:
            commands = []
            for row in rows:
                cmd = pm.get_command_string_from_row(row)
                commands.append(cmd)
            return commands
        else:
            if _return_dict:
                command_dicts = [row.__dict__ for row in rows]
                for nn, cmd_dict in enumerate(command_dicts):
                    for key, value in cmd_dict.items():
                        try:
                            command_dicts[nn][key] = json.loads(value)
                        except (json.JSONDecodeError, TypeError):
                            pass
                return command_dicts
            else:
                return [row for row in rows]
    else:
        return []



import tempfile
import shutil
from pathlib import Path
from subprocess import run, PIPE
import os
import time

from evaluation_system.api.plugin import PluginAbstract
from evaluation_system.api.parameters import (
    ParameterDictionary,
    Integer,
    Float,
    String,
    InputDirectory,
)

from evaluation_system.model.user import User
from evaluation_system.model.db import UserDB


class DummyPlugin(PluginAbstract):
    """Stub class for implementing the abstract one"""

    __short_description__ = "A dummy plugin"
    __long_description__ = ""
    __version__ = (0, 0, 0)
    __tags__ = ["foo"]
    __category__ = "statistical"
    __name__ = "DummyPlugin"
    __parameters__ = ParameterDictionary(
        Integer(name="number", help="This is just a number, not really important"),
        Integer(
            name="the_number",
            mandatory=True,
            help="This is *THE* number. Please provide it",
        ),
        String(name="something", default="test"),
        Float(name="other", default=1.4),
        InputDirectory(name="input", help="An input file"),
        String(name="variable", default="tas", help="An input variable"),
    )
    _runs = []
    _template = "${number} - $something - $other"
    tool_developer = {"name": "DummyUser", "email": "data@dkrz.de"}

    def run_tool(self, config_dict=None):
        DummyPlugin._runs.append(config_dict)
        num = config_dict.get("other", 1.4)
        print(num)
        if num < 0:
            time.sleep(-num)
        tool_path = Path(__file__).parent / "plugin_env" / "bin" / "python"
        res = run(["which", "python"], stdout=PIPE, stderr=PIPE)
        assert "plugin_env" in os.environ["PATH"]
        out = res.stdout.decode().strip()
        assert out == str(tool_path.absolute())
        print(f"Dummy tool was run with: {config_dict}")
        return {
            "/tmp/dummyfile1": dict(type="plot"),
            "/tmp/dummyfile2": dict(type="data"),
        }


class DummyUser(User):
    """Create a dummy User object that allows testing"""

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self._random_home:
            self.cleanRandomHome()
        return True

    def __init__(self, random_home=False, uid=None, **override):
        self._random_home = None
        self.username = override.get("pw_name", None)
        if random_home:
            if "pw_dir" in override:
                raise Exception("Can't define random_home and provide a home directory")
            override["pw_dir"] = tempfile.mkdtemp("_dummyUser")
            self._random_home = override["pw_dir"]
        super().__init__(uid=uid)

        class DummyUserData(list):
            """Override a normal list and make it work like the pwd read-only struct"""

            _NAMES = "pw_name pw_passwd pw_uid pw_gid pw_gecos pw_dir pw_shell".split()

            def __init__(self, arr_list):
                list.__init__(self, arr_list)

            def __getattribute__(self, name):
                # don't access any internal variable (avoid recursion!)
                if name[0] != "_" and name in self._NAMES:
                    return self[self._NAMES.index(name)]
                return list.__getattribute__(self, name)

        # copy the current data
        user_data = list(self._userdata[:])
        for key, value in override.items():
            user_data[DummyUserData._NAMES.index(key)] = value
        self._userdata = DummyUserData(user_data)
        self._db = UserDB(self)

    def cleanRandomHome(self):
        try:
            # make sure the home is a temporary one!!!
            shutil.rmtree(self._random_home)
        except (FileNotFoundError, OSError):
            pass
